/* -------------------------------------------------------------- */
/*  LAST EDIT: Sat Aug 17 18:43:19 1996 by Adabas D Database (gutamil.v12-berlin.de!adabas)  */
/*  $Id: AdabasLong.c,v 1.9 1996/08/17 16:43:25 adabas Exp $ */
/* -------------------------------------------------------------- */
/* Copyright (c) 1996 Christian Krone. All rights reserved.       */
/* This program is free software; you can redistribute it and/or  */
/* modify it under the same terms as Tcl itself.                  */
/* See also licence.terms                                         */
/* -------------------------------------------------------------- */

/* ----------------- Import interface --------------------------- */
#include <string.h>
#include <fcntl.h>
#include <tcl.h>
#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#else
   extern long lseek (int fd, long offset, int whence);
   extern int  read  (int fd, char *buf, size_t size);
   extern int  write (int fd, char *buf, size_t size);
   extern int  close (int fd);
#endif

#include "adabas.h"

#include "AdabasUtil.h"
#include "AdabasRte.h"
#include "AdabasPacket.h"
#include "AdabasSend.h"
#include "AdabasSql.h"

/* ----------------- Export interface --------------------------- */
#include "AdabasLong.h"

/* ----------------- Local constants ---------------------------- */
#if ADABAS_VERSION >= 62
#  define LONG_DESCRIPTOR    41
#  define LONG_VALPOS_FIRST  (LONG_DESCRIPTOR+1)
#  define LONG_VALPOS_NEXT   (LONG_DESCRIPTOR+1)
#  define LONG_HEADER        200
#  define LONG_SIZE_READ     MAXINT4
#  define LONG_SIZE_FIRST(i) ((i)->rteInfo.sqlPacketSize \
			      - LONG_HEADER - LONG_VALPOS_FIRST)
#  define LONG_SIZE_NEXT(i)  ((i)->rteInfo.sqlPacketSize \
			      - LONG_HEADER - LONG_VALPOS_NEXT)
#  define BUFPOS             bufpos
#else /* ADABAS_VERSION <= 61 */
#  define LONG_DESCRIPTOR   33
#  define LONG_VALPOS_FIRST  (LONG_DESCRIPTOR+11)
#  define LONG_VALPOS_NEXT   (LONG_DESCRIPTOR+13)
#  define LONG_SIZE_READ     MAXINT2
#  define LONG_SIZE_FIRST(i) (VARPART - LONG_VALPOS_FIRST - 1)
#  define LONG_SIZE_NEXT(i)  (VARPART - LONG_VALPOS_NEXT)
#  define BUFPOS             var.s.bufpos
#endif 

/* ----------------- Local procedure headings ------------------- */
static long fileLength _ANSI_ARGS_((Tcl_Interp *interp, char *fileName));
static int  readFromFile _ANSI_ARGS_((Tcl_Interp *interp, char *fileName,
				      char *data, long offset, long length));

/* ----------------- Exported procedures ------------------------- */

int AdabasLongSelect (interp, cursorVars, tableName, columnName,
		      whereCond, longDesc)
     Tcl_Interp      *interp;		/* Current interpreter. */
     CursorInfo      *cursorVars;	/* Current cursor. */
     char            *tableName;	/* Name of table with long column. */
     char            *columnName;	/* Name of long column in table. */
     char            *whereCond;	/* Where condition. */
     long_descriptor *longDesc;		/* Descriptor returned by a fetch. */
{
  Tcl_DString selCommand;
  int         ret;
  char        buffer[100];

  Tcl_DStringInit   (&selCommand);
  Tcl_DStringAppend (&selCommand, "SELECT ",  -1);
  Tcl_DStringAppend (&selCommand, columnName, -1);
  Tcl_DStringAppend (&selCommand, " FROM ",   -1);
  Tcl_DStringAppend (&selCommand, tableName,  -1);
  Tcl_DStringAppend (&selCommand, " WHERE ",  -1);
  Tcl_DStringAppend (&selCommand, whereCond,  -1);
  ret = AdabasSendCmdPacket (interp, cursorVars->logonInfo, 1, 0,
			     Tcl_DStringValue (&selCommand));
  Tcl_DStringFree (&selCommand);

  if (ret == TCL_ERROR) return TCL_ERROR;

#if ADABAS_VERSION >= 62
  if (!getResultTable (&cursorVars->logonInfo->rcvPacket->segm,
		       cursorVars->resultTable))
    *cursorVars->resultTable = 0;
#else /* ADABAS_VERSION <= 61 */
  if (!getResultTable (&cursorVars->logonInfo->sqlPacket,
		       cursorVars->resultTable))
    *cursorVars->resultTable = 0;
#endif

  sprintf (buffer, "FETCH %s INTO ?", cursorVars->resultTable);
  if (AdabasSendCmdPacket (interp, cursorVars->logonInfo,
			   1, 0, buffer) == TCL_ERROR)
    return TCL_ERROR;
#if ADABAS_VERSION >= 62
  cursorVars->cntParams =
    getShortInfos (&cursorVars->logonInfo->rcvPacket->segm,
		   &cursorVars->shortInfos);
  ret = cursorVars->cntParams == 1
    && datatypeIsLong (cursorVars->shortInfos->data_type);
#else /* ADABAS_VERSION <= 61 */
  cursorVars->cntParams = getShortInfos (&cursorVars->logonInfo->sqlPacket,
					 &cursorVars->shortInfos);
  ret = cursorVars->cntParams == 1
    && datatypeIsLong (cursorVars->shortInfos->d_type);
#endif
  if (!ret) {
    interp->result = "invalid select parameter for adareadlong";
    return TCL_ERROR;
  }
    
  return AdabasLongDescriptor (interp, cursorVars->logonInfo, 
			       longDesc, 1, longSelect);
}

/* -------------------------------------------------------------- */

int AdabasLongReadFirst (interp, cursorVars, longDesc, fileName, bytes, cont)
     Tcl_Interp      *interp;		/* Current interpreter. */
     CursorInfo      *cursorVars;	/* Current cursor. */
     long_descriptor *longDesc;		/* Descriptor returned by select. */
     char            *fileName;		/* Filename to write long into. */
     long            *bytes;		/* Bytes read in this round. */
     int             *cont;		/* Is there more data to read? */
{
  char       *data;
  int         length;
  Tcl_DString longVal;

  if (!(data = getDataOfCursor (interp, cursorVars, FOR_LONG_OP)))
    return TCL_ERROR;

  data  += longDesc->valpos - 1;
  length = longDesc->vallen;

  if (fileName) {
    if (AdabasLongToFile (interp, 1, fileName, length, data) == TCL_ERROR)
      return TCL_ERROR;
  } else {
    Tcl_DStringInit (&longVal);
    escapeSequence (&longVal, data, length);
    Tcl_DStringResult (interp, &longVal);
  }

  *bytes = length;
  *cont   = longDesc->valmode == VM_DATAPART;

  return TCL_OK;
}

/* -------------------------------------------------------------- */

int AdabasLongReadNext (interp, cursorVars, longDesc, fileName, bytes, cont)
     Tcl_Interp      *interp;		/* Current interpreter. */
     CursorInfo      *cursorVars;	/* Current cursor. */
     long_descriptor *longDesc;		/* Descriptor returned by select. */
     char            *fileName;		/* Filename to write long into. */
     long            *bytes;		/* Bytes read in this round. */
     int             *cont;		/* Is there more data to read? */
{
  char       *data;
  int         length;
  Tcl_DString longVal;

  longDesc->vallen = LONG_SIZE_READ;

  /*
   * Send a GETVAL order to the database server to get
   * the next portion of this long column.
   */
  if (AdabasSendValPacket (interp, cursorVars->logonInfo,
			   M_GETVAL, longDesc) == TCL_ERROR)
    return TCL_ERROR;

  if (!(data = getDataOfCursor (interp, cursorVars, FOR_LONG_OP)))
    return TCL_ERROR;

  data  += longDesc->valpos - 1;
  length = longDesc->vallen;

  if (fileName) {
    if (AdabasLongToFile (interp, 2, fileName, length, data) == TCL_ERROR)
      return TCL_ERROR;
  } else {
    Tcl_DStringInit   (&longVal);
    Tcl_DStringAppend (&longVal, data, length);
    Tcl_AppendResult  (interp, Tcl_DStringValue (&longVal), (char *) NULL);
    Tcl_DStringFree   (&longVal);
  }

  *bytes += length;
  *cont   = longDesc->valmode == VM_DATAPART;

  return TCL_OK;
}

/* -------------------------------------------------------------- */

int AdabasLongUpdate (interp, cursorVars, tableName, columnName,
		      whereCond, parsId)
     Tcl_Interp  *interp;		/* Current interpreter. */
     CursorInfo  *cursorVars;		/* Current cursor. */
     char        *tableName;		/* Name of table with long column. */
     char        *columnName;		/* Name of long column in table. */
     char        *whereCond;		/* Where condition. */
     Tcl_DString *parsId;		/* ParsId returned by this update. */
{
  AdabasInfo *vars = cursorVars->logonInfo;
  Tcl_DString updCommand;
  int         oldMessType;
  int         ret;

  /*
   *  Construct the 'UPDATE OF table SET column = :x WHERE condition' command
   *  and let it parse by the database kernel.
   */
  Tcl_DStringInit   (&updCommand);
  Tcl_DStringAppend (&updCommand, "UPDATE OF ",   -1);
  Tcl_DStringAppend (&updCommand, tableName,      -1);
  Tcl_DStringAppend (&updCommand, " SET ",        -1);
  Tcl_DStringAppend (&updCommand, columnName,     -1);
  Tcl_DStringAppend (&updCommand, " = :x WHERE ", -1);
  Tcl_DStringAppend (&updCommand, whereCond,      -1);
  oldMessType               = vars->packetInfo.messType;
  vars->packetInfo.messType = M_PARSE;
  ret = AdabasSendCmdPacket (interp, vars, 1, 0,
			     Tcl_DStringValue (&updCommand));
  vars->packetInfo.messType = oldMessType;
  Tcl_DStringFree (&updCommand);
  if (ret == TCL_ERROR) return TCL_ERROR;

  Tcl_DStringInit (parsId);
#if ADABAS_VERSION >= 62
  cursorVars->cntParams = getShortInfos (&vars->rcvPacket->segm,
					 &cursorVars->shortInfos);
  ret = cursorVars->cntParams == 1
    && datatypeIsLong (cursorVars->shortInfos->data_type)
      && getParsId (&vars->rcvPacket->segm, parsId);
#else /* ADABAS_VERSION <= 61 */
  cursorVars->cntParams = getShortInfos (&vars->sqlPacket,
					 &cursorVars->shortInfos);
  ret = cursorVars->cntParams == 1
    && datatypeIsLong (cursorVars->shortInfos->d_type)
      && getParsId (&vars->sqlPacket, parsId);
#endif
  if (!ret) {
    interp->result = "invalid update parameter for adawritelong";
    return TCL_ERROR;
  }

  return TCL_OK;
}

/* -------------------------------------------------------------- */

int AdabasLongWriteFirst (interp, cursorVars, longDesc, parsId, fileName,
			  longValue, length, bytes, cont)
     Tcl_Interp      *interp;		/* Current interpreter. */
     CursorInfo      *cursorVars;	/* Current cursor. */
     long_descriptor *longDesc;		/* Descriptor returned by update. */
     Tcl_DString     *parsId;		/* ParsId of previous update. */
     char            *fileName;		/* Filename to read long out of. */
     char            *longValue;	/* directly given long to write. */
     int              length;		/* length, still to write. */
     int4            *bytes;		/* Bytes written in this round. */
     int             *cont;		/* Is there more data to write? */
{
  int              partI;
  char            *data;
  AdabasInfo      *vars        = cursorVars->logonInfo;
  AdabasPacketInfo thisOptions = vars->packetInfo;
  field_info      *longSi      = cursorVars->shortInfos;
#if ADABAS_VERSION >= 62
  int4             dc1, dc2;
#endif
  if ((*cont = length > LONG_SIZE_FIRST (vars)))
    length = LONG_SIZE_FIRST (vars);

  /*
   * Filling up the long descriptor.
   */
  memset ((char *) longDesc, 0, sizeof (*longDesc));
  longDesc->valmode = *cont ? VM_DATAPART : VM_ALLDATA;
  longDesc->vallen  = length;
#if ADABAS_VERSION >= 62
  longDesc->valpos  = longSi->BUFPOS + LONG_DESCRIPTOR;
#else /* ADABAS_VERSION <= 61 */
  longDesc->valpos  = PARSID + LONG_DESCRIPTOR + 1;
#endif

  /*
   * Now write the parsid for execution into the packet, followed
   * by the first portion of the long value.
   */
  for (partI = 0; partI < PART_KINDS; partI++) {
    vars->partInfos[ partI ].bufLength = 0;
  }
  vars->partInfos[ pkParsid ].buf       = Tcl_DStringValue (parsId);
  vars->partInfos[ pkParsid ].bufLength = strlen (Tcl_DStringValue (parsId));
  vars->partInfos[ pkData   ].buf       = NULL; /* we fill up later... */
  vars->partInfos[ pkData   ].bufLength =
     longSi->BUFPOS - 1 + LONG_DESCRIPTOR + longDesc->vallen;

  thisOptions.messType = M_EXECUTE;
  if (AdabasFillPacket (interp, vars, &thisOptions) == TCL_ERROR)
    return TCL_ERROR;

#if ADABAS_VERSION >= 62
  data  = getData (&vars->sndPacket->segm, &dc1, &dc2, 0);
#else /* ADABAS_VERSION <= 61*/
  data  = vars->sqlPacket.var_part+vars->sqlPacket.part1_length;
#endif
  if (!data) {
    interp->result = "No data part???";
    return TCL_ERROR;
  }

  memset (data, 0, longSi->BUFPOS); /* defined byte. */
  memcpy (data + longSi->BUFPOS, (char *) longDesc, sizeof (*longDesc));

#if ADABAS_VERSION >= 62
  data += longDesc->valpos - 1;
#else /* ADABAS_VERSION <= 61*/
  data += LONG_DESCRIPTOR;
#endif

  if (fileName) {
    if (readFromFile (interp, fileName, data, 0, length) == TCL_ERROR)
      return TCL_ERROR;
  } else {
    memcpy (data, longValue, length);
  }

  if (AdabasSendPacket (interp, vars) == TCL_ERROR)
    return TCL_ERROR;

  if (AdabasLongDescriptor (interp, vars, longDesc, *cont,
			    longWriteFirst) == TCL_ERROR)
    return TCL_ERROR;

  *bytes = length;
  return TCL_OK;
}

/* -------------------------------------------------------------- */

extern int AdabasLongWriteNext (interp, cursorVars, longDesc, fileName,
				longValue, length, bytes, cont)
     Tcl_Interp      *interp;		/* Current interpreter. */
     CursorInfo      *cursorVars;	/* Current cursor. */
     long_descriptor *longDesc;		/* Descriptor returned by update. */
     char            *fileName;		/* Filename to read long out of. */
     char            *longValue;	/* directly given long to write. */
     int              length;		/* length, still to write. */
     int4            *bytes;		/* Bytes written in this round. */
     int             *cont;		/* Is there more data to write? */
{
  int              partI;
  int4             dc1, dc2;
  char            *data;
  AdabasInfo      *vars        = cursorVars->logonInfo;
  AdabasPacketInfo thisOptions = vars->packetInfo;
  char            *longDescPtr;
#if ADABAS_VERSION <= 61
  int2             argCount = 1;
#endif

  length -= *bytes;
  if (length < 0) {
    sprintf (interp->result,
	     "AdabasLongWriteNext: Invalid length %d???", length);
    return TCL_ERROR;
  }
  if ((*cont = length > LONG_SIZE_NEXT (vars)))
    length = LONG_SIZE_NEXT (vars);

  /*
   * Filling up the long descriptor.
   */
  longDesc->valmode = *cont ? VM_DATAPART : VM_LASTDATA;
  longDesc->vallen  = length;
  longDesc->valpos  = LONG_VALPOS_NEXT;

  /*
   * Now construct the packet of mess type PUTVAL, which consists
   * solely of a LONGDATA part.
   */
  for (partI = 0; partI < PART_KINDS; partI++) {
    vars->partInfos[ partI ].bufLength = 0;
  }
  vars->partInfos[ pkLongdata ].buf       = NULL; /* we fill up later... */
  vars->partInfos[ pkLongdata ].bufLength = 
    LONG_VALPOS_NEXT + longDesc->vallen - 1;

  thisOptions.messType = M_PUTVAL;
  if (AdabasFillPacket (interp, vars, &thisOptions) == TCL_ERROR)
    return TCL_ERROR;

#if ADABAS_VERSION >= 62
  data  = getData (&vars->sndPacket->segm, &dc1, &dc2, FOR_LONG_OP);
#else /* ADABAS_VERSION <= 61*/
  data  = getData (&vars->sqlPacket, &dc1, &dc2, FOR_LONG_OP);
#endif
  if (!data) {
    interp->result = "No data part???";
    return TCL_ERROR;
  }

#if ADABAS_VERSION >= 62
  longDescPtr = data;
#else /* ADABAS_VERSION <= 61*/
  memcpy (data, (char *) &argCount, sizeof (int2));
  longDescPtr = data + sizeof (int2);
#endif

  *longDescPtr = 0; /* defined byte. */
  memcpy (longDescPtr+1, (char *) longDesc, sizeof (*longDesc));

  data += longDesc->valpos - 1;

  if (fileName) {
    if (readFromFile (interp, fileName, data, *bytes, length) == TCL_ERROR)
      return TCL_ERROR;
  } else {
    memcpy (data, longValue + *bytes, length);
  }

  if (AdabasSendPacket (interp, vars) == TCL_ERROR)
    return TCL_ERROR;

  if (AdabasLongDescriptor (interp, vars, longDesc, *cont,
			    longWriteNext) == TCL_ERROR)
    return TCL_ERROR;

  *bytes += length;
  return TCL_OK;
}

/* -------------------------------------------------------------- */

int AdabasLongWriteEnd (interp, cursorVars, longDesc)
     Tcl_Interp      *interp;		/* Current interpreter. */
     CursorInfo      *cursorVars;	/* Current cursor. */
     long_descriptor *longDesc;		/* Descriptor to close. */
{
  AdabasInfo      *vars        = cursorVars->logonInfo;
  AdabasPacketInfo thisOptions = vars->packetInfo;
  int              partI;

  longDesc->valmode = VM_LAST_PUTVAL;

  /*
   * Now write the long descriptor into the packet.
   */
  for (partI = 0; partI < PART_KINDS; partI++) {
    vars->partInfos[ partI ].bufLength = 0;
  }
  vars->partInfos[ pkData   ].buf       = NULL; /* we fill up later... */
  vars->partInfos[ pkData   ].bufLength = 1 + LONG_DESCRIPTOR;

  thisOptions.messType = M_PUTVAL;
  if (AdabasFillPacket (interp, vars, &thisOptions) == TCL_ERROR)
    return TCL_ERROR;

  return AdabasSendPacket (interp, vars);
}

/* -------------------------------------------------------------- */

int AdabasLongToFile (interp, toFile, fileName, length, data)
     Tcl_Interp *interp;		/* Current interpreter. */
     int         toFile;		/* File mode: 1=create; 2=append. */
     char       *fileName;		/* Filename to write data into. */
     int         length;		/* Length to write. */
     char       *data;			/* Long data to write. */
{
  int file;
  int mode;

  switch (toFile) {
  case 2:  mode = O_WRONLY | O_APPEND;          break;
  case 1:  mode = O_WRONLY | O_TRUNC | O_CREAT; break;
  default: return TCL_ERROR;
  }

  if ((file = open (fileName, mode, 0666)) < 0) {
    Tcl_AppendResult (interp, fileName, "\": ",
		      Tcl_PosixError (interp), (char *) NULL);
    return TCL_ERROR;
  }
  write (file, data, length);
  close (file);

  return TCL_OK;
}

/* -------------------------------------------------------------- */

int AdabasLongDescriptor (interp, vars, longDesc, isError, action)
     Tcl_Interp      *interp;		/* Current interpreter. */
     AdabasInfo      *vars;		/* Current logon handle. */
     long_descriptor *longDesc;		/* Returned long descriptor. */
     int              isError;		/* Is it an error to find no descr? */
     LongAction       action;		/* For whom we look for a descr. */
{
  char *longDescPtr;
  char *procName;

#if ADABAS_VERSION >= 62
  longDescPtr = getLongdesc (&vars->rcvPacket->segm);
#else /* ADABAS_VERSION <= 61 */
  longDescPtr = getLongdesc (&vars->sqlPacket, action != longSelect);
#endif

  if (longDescPtr) {
    memcpy (longDesc, longDescPtr, sizeof (*longDesc));
  } else if (isError) {
    switch (action) {
    case longSelect:     procName = "longSelect";     break;
    case longWriteFirst: procName = "longWriteFirst"; break;
    case longWriteNext:  procName = "longWriteNext";  break;
    case longSendVal:    procName = "longSendVal";    break;
    default:             procName = "unknown action"; break;
    }
    sprintf (interp->result, "%s: No returning long descriptor???", procName);
    return TCL_ERROR;
  }
  return TCL_OK;
}

/* -------------------------------------------------------------- */

int AdabasLongValLength (interp, fileName, longValue, substVal, length)
     Tcl_Interp  *interp;		/* Current interpreter. */
     char        *fileName;		/* Filename to read long out of. */
     char        *longValue;		/* directly given long to read. */
     Tcl_DString *substVal;		/* Substituted value of longValue,
					 * if longValue is not NULL. */
     int4        *length;		/* returning length of file or val. */
{
  if (fileName) {
    *length = fileLength (interp, fileName);
    Tcl_DStringInit (substVal);
  } else {
    *length = substBackslash (substVal, longValue, 1, strlen (longValue));
  }
  if (*length < 0) return TCL_ERROR;
  return TCL_OK;
}


/* ----------------- Local procedures --------------------------- */

static long fileLength (interp, fileName)
     Tcl_Interp  *interp;		/* Current interpreter. */
     char        *fileName;		/* Filename to read long out of. */
{
  int  file;
  long length;

  if ((file = open (fileName, O_RDONLY)) < 0) {
    Tcl_AppendResult (interp, fileName, ": ",
		      Tcl_PosixError (interp), (char *) NULL);
    return -1;
  }

  length = lseek (file, 0, 2);
  close (file);

  return length;
}

static int readFromFile (interp, fileName, data, offset, length)
     Tcl_Interp *interp;		/* Current interpreter. */
     char       *fileName;		/* Filename to read long out of. */
     char       *data;			/* Buffer to read long data into. */
     long        offset;		/* Where to start the reading. */
     long        length;		/* Length to read. */
{
  int  file;
  long ret;

  if ((file = open (fileName, O_RDONLY)) < 0) {
    Tcl_AppendResult (interp, fileName, ": ",
		      Tcl_PosixError (interp), (char *) NULL);
    return TCL_ERROR;
  }

  ret = lseek (file, offset, 0);
  if (ret != offset) {
    sprintf (interp->result, "seek to position %ld failed", offset);
    return TCL_ERROR;
  }
  
  ret = read (file, data, length);
  close (file);

  if (ret != length) {
    sprintf (interp->result, "wanted %ld bytes, got only %ld", length, ret);
    return TCL_ERROR;
  }

  return TCL_OK;
}
