/* -------------------------------------------------------------- */
/*  LAST EDIT: Sat Aug 17 18:44:43 1996 by Adabas D Database (gutamil.v12-berlin.de!adabas)  */
/*  $Id: AdabasRte.c,v 1.24 1996/08/17 16:44:51 adabas Exp $ */
/* -------------------------------------------------------------- */
/* Copyright (c) 1996 Christian Krone. All rights reserved.       */
/* This program is free software; you can redistribute it and/or  */
/* modify it under the same terms as Tcl itself.                  */
/* See also licence.terms                                         */
/* -------------------------------------------------------------- */

/* ----------------- Import interface --------------------------- */
#include <stdlib.h>
#include <string.h>
#include <tcl.h>

#include "adabas.h"
#include "sqlrte.h"

#include "AdabasUtil.h"

/* ----------------- Export interface --------------------------- */
#include "AdabasRte.h"

/* ----------------- Exported procedures ------------------------- */

void AdabasRteInit (vars)
     AdabasRteInfo *vars;		/* Current RunTimeEnvironment. */
{
  c64        component;

  /*
   * A call of 'sqlinit' to initialize the rte.
   */
  c2p (component,  "ADATCL", sizeof (component));
  vars->cancel = 0;
  sqlinit (component, &vars->cancel);

#if ADABAS_VERSION <= 61
  /*
   * In pre6.2 the size of a packet is rather static.
   * So here we can wrote it once for all times.
   */
  vars->sqlPacketSize = sizeof (varpart);
#endif
}

/* -------------------------------------------------------------- */

int AdabasRteConnect (interp, vars, serverdb, serv)
     Tcl_Interp    *interp;		/* Current interpreter. */
     AdabasRteInfo *vars;		/* Current RunTimeEnvironment. */
     char          *serverdb;		/* Serverdb to connect to. */
     int            serv;		/* Service to open (USER/UTILITY). */
{
  comm_error returncode;
  errtext    errtext;
  char       c_errtext[ERRTEXT+1];

  /*
   * Open the connection to the given database with 'sqlaconnect'.
   */
  c2p (vars->dbnode, "",       sizeof (vars->dbnode));
  c2p (vars->dbname, serverdb, sizeof (vars->dbname));
  vars->reference = 0;
  vars->service   = serv;
  sqlaconnect (0, vars->dbnode, vars->dbname, vars->service,
#if ADABAS_VERSION >= 62
	       1, &vars->reference, &vars->sqlPacketSize, vars->sqlPacketList,
#else /* ADABAS_VERSION <= 61 */
	       &vars->reference, 
#endif
	       errtext, &returncode);
  if (returncode) {
    p2c (c_errtext, errtext, sizeof (errtext));
    Tcl_AppendResult (interp, "connect to \"", serverdb, "\" failed: ",
		      c_errtext, (char *) NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

/* -------------------------------------------------------------- */

int AdabasRteSend (interp, vars, snd_packet
#if ADABAS_VERSION >= 62
		   , rcv_packet
#endif
                   )
     Tcl_Interp    *interp;		/* Current interpreter. */
     AdabasRteInfo *vars;		/* Current RunTimeEnvironment. */
     packet        *snd_packet;		/* Packet to send to the kernel. */
#if ADABAS_VERSION >= 62
     packet      **rcv_packet;		/* Packet received from the kernel. */
#endif
{
#if ADABAS_VERSION >= 62
  int4       packet_length;
  int4       res_packet_length;
#else /* ADABAS_VERSION <= 61 */
  int2       packet_length;
  int2       res_packet_length;
#endif
  errtext    errtext;
  char       c_errtext[ERRTEXT+1];
  comm_error returncode;

  /*
   * Send the constructed packet to the database with 'sqlarequest'.
   * but first of all evaluate the packet length.
   */
#if ADABAS_VERSION >= 62
  packet_length = snd_packet->header.varpart_len + sizeof (packet_header);
#else /* ADABAS_VERSION <= 61 */
  packet_length = PACKETHEAD
    + snd_packet->part1_length + snd_packet->part2_length;
#endif
  sqlarequest (vars->reference, snd_packet, packet_length,
	       errtext, &returncode);
  if (returncode) {
    p2c (c_errtext, errtext, sizeof (errtext));
    Tcl_AppendResult (interp, "request failed: ", c_errtext, (char *) NULL);
    return TCL_ERROR;
  }

  /*
   * Now wait for the answer of the database with 'sqlareceive'.
   */
  sqlareceive (vars->reference,
#if ADABAS_VERSION >= 62
	       (void **) rcv_packet,
#else /* ADABAS_VERSION <= 61 */
	       snd_packet,
#endif
	       &res_packet_length, errtext, &returncode);
  if (returncode) {
    p2c (c_errtext, errtext, sizeof (errtext));
    Tcl_AppendResult (interp, "receive failed: ", c_errtext, (char *) NULL);
    return TCL_ERROR;
  }
  
  return TCL_OK;
}


/* -------------------------------------------------------------- */

void AdabasRteTermId (result)
     char *result;			/* Returning terminalId. */
{
  termid terminalid;

  sqltermid (terminalid);
  sprintf (result, "%-*.*s",
	   sizeof (terminalid), (int) sizeof (terminalid), terminalid);
}

/* -------------------------------------------------------------- */

void AdabasRteVersion (rteVersion)
     char *rteVersion;			/* Returning version of RTE. */
{
  sprintf (rteVersion, "%-.*s", (int) sizeof (SQLRTEVERSION), SQLRTEVERSION);
}

/* -------------------------------------------------------------- */

void AdabasRteCancel (vars)
     AdabasRteInfo *vars;		/* Current RunTimeEnvironment. */
{
  /*
   * Cancels the execution of the currently running sql statement
   * with 'sqlacancel'.
   */
  sqlacancel (vars->reference);
}

/* -------------------------------------------------------------- */

void AdabasRteClose (vars)
     AdabasRteInfo *vars;		/* Current RunTimeEnvironment. */
{
  /*
   * Close the connection to the database with 'sqlarelease'.
   */
  sqlarelease (vars->reference);
}

/* -------------------------------------------------------------- */

int AdabasXUserClear (interp)
     Tcl_Interp    *interp;		/* Current interpreter. */
{
  if (!sqlclearuser ()) {
    interp->result = "clear of xuser file failed";
    return TCL_ERROR;
  }
  return TCL_OK;
}

/* -------------------------------------------------------------- */

int AdabasXUserOpen (interp)
     Tcl_Interp    *interp;		/* Current interpreter. */
{
  errtext errtext;
  boolean ok;
  char    c_errtext[ERRTEXT+1];

  sqlxuopenuser  (errtext, &ok);
  if (!ok) {
    p2c (c_errtext, errtext, sizeof (errtext));
    Tcl_AppendResult (interp, "open of xuser file failed: ",
		      c_errtext, (char *) NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

/* -------------------------------------------------------------- */

int AdabasXUserClose (interp)
     Tcl_Interp    *interp;		/* Current interpreter. */
{
  errtext errtext;
  boolean ok;
  char    c_errtext[ERRTEXT+1];

  sqlxucloseuser (errtext, &ok);
  if (!ok) {
    p2c (c_errtext, errtext, sizeof (errtext));
    Tcl_AppendResult (interp, "close of xuser file failed: ",
		      c_errtext, (char *) NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

/* -------------------------------------------------------------- */

int AdabasXUserIndex (interp, userIndex, userParams)
     Tcl_Interp   *interp;		/* Current interpreter. */
     int           userIndex;		/* index of user to get. */
     xuser_record *userParams;		/* returned user record. */
{
  errtext errtext;
  boolean ok;
  char    c_errtext[ERRTEXT+1];

  sqlindexuser (userIndex, userParams, errtext, &ok);
  if (!ok) {
    p2c (c_errtext, errtext, sizeof (errtext));
    Tcl_AppendResult (interp, "index of xuser file failed: ",
		      c_errtext, (char *) NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

/* -------------------------------------------------------------- */

int AdabasXUserPut (interp, userParams)
     Tcl_Interp   *interp;		/* Current interpreter. */
     xuser_record *userParams;		/* user record to write. */
{
  errtext errtext;
  boolean ok;
  char    c_errtext[ERRTEXT+1];

  sqlputuser (userParams, errtext, &ok);
  if (!ok) {
    p2c (c_errtext, errtext, sizeof (errtext));
    Tcl_AppendResult (interp, "writing of xuser file failed: ",
		      c_errtext, (char *) NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

/* -------------------------------------------------------------- */

int AdabasXUserGet (interp, userKey, userParams)
     Tcl_Interp   *interp;		/* Current interpreter. */
     char         *userKey;		/* User key, to whom record belongs. */
     xuser_record *userParams;		/* returned user record. */
{
  errtext errtext;
  boolean ok;
  char    c_errtext[ERRTEXT+1];

  /*
   * Get the user data of the given XUSER key.
   */
  c2p (userParams->xu_key, userKey ? userKey : "DEFAULT",
       sizeof (userParams->xu_key));
  sqlgetuser (userParams, errtext, &ok);
  if (!ok) {
    p2c (c_errtext, errtext, sizeof (errtext));
    Tcl_AppendResult (interp, "xuser of ", userKey ? userKey : "default user",
		      " failed: ", c_errtext, (char *) NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

/* -------------------------------------------------------------- */

int AdabasXUserArgs (interp, userParams, givenPasswd)
     Tcl_Interp   *interp;		/* Current interpreter. */
     xuser_record *userParams;		/* returned user record. */
     char         *givenPasswd;		/* password from command line.*/
{
  SQL_PASSWORD        password;
  struct args_options options;
  xuser_set           xusertype;
  c40                 errtext;
  boolean             ok;
  char                c_errtext[ERRTEXT+1];
  int                 orgArgc;
  int                *orgArgv;
  int                 ix;
  char                evalString[80];

  /*
   * We must put the arguments into the global variables argc and argv
   * outselves. And we better store the original...
   */
  strcpy (evalString, "linsert $argv 0 $argv0");
  if (Tcl_GlobalEval (interp, evalString) == TCL_ERROR) {
    interp->result = "Variable \"argv\" or \"argv0\" not set";
    return TCL_ERROR;
  }
  orgArgc = _argc;
  orgArgv = _argv;
  Tcl_SplitList (interp, interp->result, &_argc, (char ***)&_argv);
  Tcl_ResetResult (interp);

  memset (userParams, ' ', sizeof (*userParams));
  *givenPasswd = '\0';

  /*
   * Get the user data, that was specified on the command line,
   * in the environment variables or in the xuser default entry.
   */
  options.opt_component = SQL_QUERY;
  sqlarg3 (userParams, password, &options, xusertype, errtext, &ok);

  /*
   * Do anything unhappen on the global level.
   */
  free ((char *) _argv);
  _argc = orgArgc;
  _argv = orgArgv;

  if (!ok) {
    p2c (c_errtext, errtext, sizeof (errtext));
    Tcl_AppendResult (interp, "xuser of program arguments failed: ",
		      c_errtext, (char *) NULL);
    return TCL_ERROR;
  }

  /*
   * If a password was given on the command line (with -u user,passwd)
   * sqlarg3 gives it via its parameter password and we have to give
   * it to the caller ourself.
   */
  for (ix = 0; ix < sizeof (password); ix++) {
    if (password[ix] != ' ') {
      p2c (givenPasswd, password, sizeof (password));
      ix = sizeof (password); /* terminate loop. */
    }
  }

  return TCL_OK;
}
