/* -------------------------------------------------------------- */
/*  LAST EDIT: Sun Aug 18 12:39:45 1996 by Adabas D Database (gutamil.v12-berlin.de!adabas)  */
/*  $Id: AdabasSql.c,v 1.32 1996/08/18 10:45:44 adabas Exp $ */
/* -------------------------------------------------------------- */
/* Copyright (c) 1996 Christian Krone. All rights reserved.       */
/* This program is free software; you can redistribute it and/or  */
/* modify it under the same terms as Tcl itself.                  */
/* See also licence.terms                                         */
/* -------------------------------------------------------------- */

/* ----------------- Import interface --------------------------- */
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <tcl.h>

#include "adabas.h"

#include "AdabasRte.h"
#include "AdabasPacket.h"
#include "AdabasFormat.h"
#include "AdabasSend.h"
#include "AdabasUtil.h"
#include "AdabasSql.h"
#include "AdabasLong.h"

/* ----------------- Local constants ---------------------------- */
#define MAX_CURSORS  25

#define USER_RANGE   1
#define PASSWD_RANGE 2

#define CMD_RANGE    1

/* ----------------- Type definitions --------------------------- */
typedef struct {
  char        *name;
  Tcl_CmdProc *proc;
} ProcTable;

typedef enum {
  noSelect,
  massSelect,
  intoSelect,
  showSelect,
  explainSelect
} SelectKind;

/* ----------------- Local procedure headings ------------------- */
static int AdaLogonCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				    int argc, char **argv));
static int AdaLogoffCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				     int argc, char **argv));
static int AdaOpenCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				   int argc, char **argv));
static int AdaCloseCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				    int argc, char **argv));
static int AdaSqlCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				  int argc, char **argv));
static int AdaCancelCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				     int argc, char **argv));
static int AdaFetchCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				    int argc, char **argv));
static int AdaColsCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				   int argc, char **argv));
static int AdaReadLongCmd _ANSI_ARGS_((ClientData clientData,
				       Tcl_Interp *interp,
				       int argc, char **argv));
static int AdaWriteLongCmd _ANSI_ARGS_((ClientData clientData,
					Tcl_Interp *interp,
					int argc, char **argv));
static int AdaAutoCommitCmd _ANSI_ARGS_((ClientData clientData,
					 Tcl_Interp *interp,
					 int argc, char **argv));
static int AdaCommitCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				     int argc, char **argv));
static int AdaRollbackCmd _ANSI_ARGS_((ClientData clientData,
				       Tcl_Interp *interp,
				       int argc, char **argv));
static int AdaUtilCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				   int argc, char **argv));
static int AdaFetchAll _ANSI_ARGS_((Tcl_Interp *interp, char *cursorName,
				    int count, char *command));
static Tcl_HashEntry *getCursorHandle _ANSI_ARGS_((Tcl_Interp *interp,
						   char *cursorName,
						   CursorInfo **cursorVars));
static int getLogonHandle _ANSI_ARGS_((Tcl_Interp *interp, char *logonName,
				       AdabasInfo **logon));
static int  selectInto _ANSI_ARGS_((Tcl_Interp *interp, CursorInfo *cursorVars,
				    char *selectInto));
static void examineSqlCommand _ANSI_ARGS_((Tcl_Interp *interp, char *cmdString,
					   int *withInfo, int *withResCount,
					   SelectKind *selectKind,
					   char **selectInto));
static char *getNextData _ANSI_ARGS_((Tcl_Interp *interp, char *data,
				      CursorInfo *cursorVars));
static int appendOneValue _ANSI_ARGS_((Tcl_Interp *interp, char *data,
				       char *nullValue, field_info *currSi,
				       int *longColFound, int isElement));
static char *getNullValue _ANSI_ARGS_((Tcl_Interp *interp));
static CursorInfo *allocCursor _ANSI_ARGS_((Tcl_Interp *interp,
					    AdabasInfo *vars));
static void freeCursor _ANSI_ARGS_((CursorInfo *cursorVars));

/* ----------------- Local variables ---------------------------- */
static ProcTable adabasProcs[] = {
  { "adalogon",     AdaLogonCmd       },
  { "adalogoff",    AdaLogoffCmd      },
  { "adaopen",      AdaOpenCmd        },
  { "adaclose",     AdaCloseCmd       },
  { "adasql",       AdaSqlCmd        },
  { "adacancel",    AdaCancelCmd      },
  { "adafetch",     AdaFetchCmd       },
  { "adacols",      AdaColsCmd        },
  { "adareadlong",  AdaReadLongCmd    },
  { "adawritelong", AdaWriteLongCmd   },
  { "adaautocom",   AdaAutoCommitCmd  },
  { "adacommit",    AdaCommitCmd      },
  { "adarollback",  AdaRollbackCmd    },
  { "adautil",      AdaUtilCmd        },
  { (char *) NULL, (Tcl_CmdProc *) NULL }
};

static Tcl_HashTable cursorHash;

/* ----------------- Exported procedures ------------------------- */

int AdabasSqlInit (interp)
     Tcl_Interp *interp;		/* Current interpreter. */
{
  ProcTable  *p;
  Tcl_DString fileName;
  int         erg;
  char        version[200];
  char       *adabasTclDir;

  Tcl_InitHashTable (&cursorHash, TCL_STRING_KEYS);

  for (p = adabasProcs; p->name; p++) {
    Tcl_CreateCommand (interp, p->name, p->proc,
		       (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  }

  SET_VERSION (version);
  Tcl_SetVar2 (interp, ADABAS_MESSAGES, "version",   version, TCL_GLOBAL_ONLY);
  Tcl_SetVar2 (interp, ADABAS_MESSAGES, "nullvalue", "",      TCL_GLOBAL_ONLY);
  adabasTclDir = Tcl_GetVar (interp,    "adabastcl_library",  TCL_GLOBAL_ONLY);

  Tcl_DStringInit (&fileName);
  Tcl_DStringAppend (&fileName, adabasTclDir,  -1);
  Tcl_DStringAppend (&fileName, "/adasql.tcl", -1);
  erg = Tcl_EvalFile (interp, Tcl_DStringValue (&fileName));
  Tcl_DStringFree (&fileName);
  return erg;
}

/* -------------------------------------------------------------- */

char *getDataOfCursor (interp, cursorVars, flags)
     Tcl_Interp *interp;		/* Current interpreter. */
     CursorInfo *cursorVars;		/* Current cursor. */
     int         flags;			/* defined in AdabasPacket.h */
{
  char *data;

#if ADABAS_VERSION >= 62
  data = getData (&cursorVars->logonInfo->rcvPacket->segm,
		  &cursorVars->dataLength, &cursorVars->massRows, flags);

#else /* ADABAS_VERSION <= 61 */
  data = getData (&cursorVars->logonInfo->sqlPacket,
		  &cursorVars->dataLength, &cursorVars->massRows, flags);
#endif

  cursorVars->rowIx = 0;

  if (!data) {
    interp->result = "Oops, no data part found???";
    return (char *) NULL;
  }
  return data;
}

/* ----------------- Local procedures --------------------------- */

static int AdaOpenCmd (dummy, interp, argc, argv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         argc;			/* Number of arguments. */
     char      **argv;			/* Argument strings. */
{
  AdabasInfo    *vars;
  Tcl_HashEntry *newEntry;
  int            added;
  int            cnt;
  char           cursorName[20];
  CursorInfo    *cursorVars;

  if (argc != 2) {
    Tcl_AppendResult (interp, "wrong # args: should be \"", argv[0],
		      " logonHandle\"", (char *) NULL);
    return TCL_ERROR;
  }

  if (getLogonHandle (interp, argv[1], &vars) == TCL_ERROR)
    return TCL_ERROR;

  if (vars->packetInfo.messType != M_DBS) {
    Tcl_AppendResult (interp, "Logon handle \"", argv[1],
		      "\" must be user service", (char *) NULL);
    return TCL_ERROR;
  }
			 
  for (cnt = 1; cnt <= MAX_CURSORS; cnt++) {
    sprintf (cursorName, "cursor%d", cnt);
    newEntry = Tcl_CreateHashEntry (&cursorHash, cursorName, &added);
    if (added) break;
  }
  if (cnt > 25) {
    sprintf (interp->result,
	     "All %d cursor handles are given away", MAX_CURSORS);
    return TCL_ERROR;
  }

  if (!(cursorVars = allocCursor (interp, vars)))
    return TCL_ERROR;

  Tcl_SetHashValue (newEntry, cursorVars);

  Tcl_SetVar2 (interp, ADABAS_MESSAGES, "handle",
	       cursorName, TCL_GLOBAL_ONLY);
  Tcl_SetResult (interp, cursorName, TCL_VOLATILE);
  return TCL_OK;
}

/* -------------------------------------------------------------- */

static int AdaCloseCmd (dummy, interp, argc, argv)
     ClientData dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int argc;				/* Number of arguments. */
     char **argv;			/* Argument strings. */
{
  CursorInfo    *cursorVars;
  Tcl_HashEntry *cursorHandle;

  if (argc != 2) {
    Tcl_AppendResult (interp, "wrong # args: should be \"", argv[0],
		      " cursorHandle\"", (char *) NULL);
    return TCL_ERROR;
  }

  cursorHandle = getCursorHandle (interp, argv[1], &cursorVars);
  if (!cursorHandle) return TCL_ERROR;

  freeCursor (cursorVars);
  Tcl_DeleteHashEntry (cursorHandle);
  return TCL_OK;
}

/* -------------------------------------------------------------- */

static CursorInfo *allocCursor (interp, vars)
     Tcl_Interp *interp;		/* Current interpreter. */
     AdabasInfo *vars;			/* Current logon. */
{
  CursorInfo *cursorVars;

  if ((cursorVars = (CursorInfo *) ckalloc (sizeof (CursorInfo))) == NULL) {
    interp->result = "Allocation of cursor failed due to lack of memory";
    return (CursorInfo *) NULL;
  }
  cursorVars->logonInfo    = vars;
  cursorVars->cntParams    = 0;
  cursorVars->longColFound = 0;
  cursorVars->dataLength   = 0;
  cursorVars->dataEnd      = 0;
  cursorVars->shortInfos   = NULL;
#if ADABAS_VERSION >= 62
  cursorVars->columnNames  = NULL;
#endif

  return cursorVars;
}

/* -------------------------------------------------------------- */

static void freeCursor (cursorVars)
     CursorInfo *cursorVars;		/* Cursor to be freed. */
{
#if ADABAS_VERSION >= 62
  if (cursorVars->columnNames) ckfree ((char *) cursorVars->columnNames);
#endif
  if (cursorVars->shortInfos)  ckfree ((char *) cursorVars->shortInfos);
  ckfree ((char *) cursorVars);
}

/* -------------------------------------------------------------- */

static int AdaSqlCmd (dummy, interp, argc, argv)
     ClientData dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int argc;				/* Number of arguments. */
     char **argv;			/* Argument strings. */
{
  CursorInfo *cursorVars;
  int         withInfo;
  int         withResCount;
  char       *selInto;
  char       *cmd;
  char       *cursorName;
  int         resCnt;
  int         ret;
  int         errorOffset = 0;
  char       *posStr;
  int         pos;
  char        numBuf[20];
  SelectKind  selectKind;
  Tcl_DString cmdStr;
#if ADABAS_VERSION >= 62
  segment    *rcvSegment;
#else /* ADABAS_VERSION <= 61 */
  Tcl_RegExp  regExp;
  char       *start;
  char       *end;
#endif

  if (argc != 3) {
    Tcl_AppendResult (interp, "wrong # args: should be \"", argv[0],
		      " cursorHandle sqlStatement\"", (char *) NULL);
    return TCL_ERROR;
  }

  cursorName = argv[1];
  if (!getCursorHandle (interp, cursorName, &cursorVars)) return TCL_ERROR;
  cursorVars->longColFound = 0;
  cursorVars->dataLength   = 0;

  /*
   * Note, that adabas doesn't like Newlines in the command...
   */
  for (cmd = argv[2]; *cmd; cmd++) if (*cmd == '\n') *cmd = ' ';
  cmd = argv[2];

  examineSqlCommand (interp, cmd, &withInfo, &withResCount,
		     &selectKind, &selInto);

  if (selectKind == massSelect) {
    Tcl_DStringInit   (&cmdStr);
    Tcl_DStringAppend (&cmdStr,  "DECLARE ",    -1);
    Tcl_DStringAppend (&cmdStr,   argv[1],      -1);
    Tcl_DStringAppend (&cmdStr, " CURSOR FOR ", -1);
    errorOffset = strlen (Tcl_DStringValue (&cmdStr));

    Tcl_DStringAppend (&cmdStr,   cmd,          -1);
    cmd = Tcl_DStringValue (&cmdStr);
  }

  ret = AdabasSendCmdPacket (interp, cursorVars->logonInfo,
			     withInfo, 1, cmd);

  if (selectKind == massSelect) Tcl_DStringFree (&cmdStr);

  if (ret == TCL_ERROR) {
    /*
     * If we modified the sql command (put a DECLARE ... CURSOR FOR
     * in front of it), we have to adjust the error position.
     */
    if (errorOffset
	&& (posStr = Tcl_GetVar2 (interp, ADABAS_MESSAGES, "errorpos",
				  TCL_GLOBAL_ONLY))
	&& Tcl_GetInt (interp, posStr, &pos) == TCL_OK
	&& pos > errorOffset) {
      sprintf (numBuf, "%d", pos - errorOffset);
      Tcl_SetVar2 (interp, ADABAS_MESSAGES, "errorpos", numBuf,
		   TCL_GLOBAL_ONLY);
    }
    return TCL_ERROR;
  }

#if ADABAS_VERSION >= 62
  rcvSegment = &cursorVars->logonInfo->rcvPacket->segm;
#endif

  /*
   * Set the resultcount, if wanted (update, delete or insert).
   */
  if (withResCount) {
#if ADABAS_VERSION >= 62
    resCnt = getResultCount (rcvSegment);
#else /* ADABAS_VERSION <= 61 */
    resCnt = getResultCount (&cursorVars->logonInfo->sqlPacket);
#endif
  } else
    resCnt = selectKind == intoSelect ? 1 : 0;
  sprintf (numBuf, "%d", resCnt);
  Tcl_SetVar2   (interp, ADABAS_MESSAGES, "rows", numBuf, TCL_GLOBAL_ONLY);
  Tcl_UnsetVar2 (interp, ADABAS_MESSAGES, "intoVars", TCL_GLOBAL_ONLY);

  if (withInfo) {
#if ADABAS_VERSION >= 62
    cursorVars->cntParams = getShortInfos (rcvSegment,
					   &cursorVars->shortInfos);
    (void) getColumnNames (rcvSegment, &cursorVars->columnNames);
#else /* ADABAS_VERSION <= 61 */
    cursorVars->cntParams = getShortInfos (&cursorVars->logonInfo->sqlPacket,
					   &cursorVars->shortInfos);
#endif

#if ADABAS_VERSION >= 62
    if (!getResultTable (rcvSegment, cursorVars->resultTable))
      *cursorVars->resultTable = 0;
#else /* ADABAS_VERSION <= 61 */
    if (!getResultTable (&cursorVars->logonInfo->sqlPacket,
			 cursorVars->resultTable))
      if (selectKind == showSelect) {
	regExp = Tcl_RegExpCompile (interp,
				    "^[ \t\n]*[a-zA-Z]*[ \t\n]*\\(([^)]*)\\)");
	if (Tcl_RegExpExec (interp, regExp, argv[2], argv[2])) {
	  Tcl_RegExpRange (regExp, 1, &start, &end);
	  if (end-start > 40) end = start+40;
	  p2c (cursorVars->resultTable, start, end-start);
	  for (start = cursorVars->resultTable; *start; start++)
	    *start = toupper (*start);
	} else {
	  strcpy (cursorVars->resultTable, "SHOW");
	}
      } else
	strcpy (cursorVars->resultTable, cursorName);
#endif
  } else 
    cursorVars->cntParams = 0;

  if (selectKind == intoSelect)
    selectInto (interp, cursorVars, selInto);

  return TCL_OK;
}

/* -------------------------------------------------------------- */

static int selectInto (interp, cursorVars, selectInto)
     Tcl_Interp *interp;		/* Current interpreter. */
     CursorInfo *cursorVars;		/* Current cursor. */
     char       *selectInto;		/* Pointing to first parameter. */
{
  Tcl_DString colName;
  Tcl_RegExp  regExp;
  char       *start;
  char       *end;
  char       *data;
  char       *nullValue;
  char       *colon  = selectInto;
  int         paramI = 0;
  int         ret    = TCL_OK;
  field_info *currSi;

  if (!(data = getDataOfCursor (interp, cursorVars, 0)))
    return TCL_ERROR;

  Tcl_DStringInit (&colName);
  regExp = Tcl_RegExpCompile (interp, ":([^, ]*)");

  currSi    = cursorVars->shortInfos;
  nullValue = getNullValue (interp);

  while (colon && Tcl_RegExpExec (interp, regExp, colon, selectInto)) {
    Tcl_RegExpRange (regExp, 1, &start, &end);
    Tcl_DStringFree   (&colName);
    Tcl_DStringAppend (&colName, start, end-start);

    ret = appendOneValue (interp, data, nullValue, currSi,
			  &cursorVars->longColFound, 0);
    Tcl_SetVar (interp, Tcl_DStringValue (&colName), interp->result, 0);
    Tcl_ResetResult (interp);

    Tcl_SetVar2 (interp, ADABAS_MESSAGES, "intoVars",
		 Tcl_DStringValue (&colName),
		 TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);

    colon = (char *) NULL;
    if (++paramI < cursorVars->cntParams) {
      while (*end == ' ') end++;
      if (*end == ',') {
	end++; 
	while (*end == ' ') end++;
	if (*end == ':') {
	  colon = end;
	  currSi++;
	}
      }
    }
  }

  Tcl_DStringFree (&colName);
  return ret;
}

/* -------------------------------------------------------------- */

static int AdaCancelCmd (dummy, interp, argc, argv)
     ClientData dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int argc;				/* Number of arguments. */
     char **argv;			/* Argument strings. */
{
  CursorInfo *cursorVars;

  if (argc != 2) {
    Tcl_AppendResult (interp, "wrong # args: should be \"", argv[0],
		      " cursorHandle\"", (char *) NULL);
    return TCL_ERROR;
  }

  if (!getCursorHandle (interp, argv[1], &cursorVars)) return TCL_ERROR;

  cursorVars->cntParams    = 0;
  cursorVars->longColFound = 0;
  AdabasRteCancel (&cursorVars->logonInfo->rteInfo);

  return TCL_OK;
}

/* -------------------------------------------------------------- */

static int AdaFetchCmd (dummy, interp, argc, argv)
     ClientData dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int argc;				/* Number of arguments. */
     char **argv;			/* Argument strings. */
{
  CursorInfo *cursorVars;
  char       *cursorName;
  char        buf[20];
  char        cmdBuf[100];
  int         length;
  int         c;
  int         ix;
  int         pos;
  char       *data;
  char       *nullValue;
  int         ret;
  field_info *currSi;
  int         massCnt = 0;
  char       *target  = "";
  char       *command = (char *) NULL;

  if (argc < 2) {
    Tcl_AppendResult (interp, "wrong # args: should be \"", argv[0],
		      " cursorHandle ?options?\"", (char *) NULL);
    return TCL_ERROR;
  }

  cursorName = argv[1];
  if (!getCursorHandle (interp, cursorName, &cursorVars)) return TCL_ERROR;

  if (!cursorVars->cntParams) {
    Tcl_AppendResult (interp, "\"", cursorName,
		      "\" not opened for fetch", (char *) NULL);
    return TCL_ERROR;
  }
  cursorVars->dataEnd = 0;

  for (argv+=2, argc-=2; argc > 1; argv += 2, argc -= 2) {
    c      = argv[0][1];
    length = strlen (argv[0]);

    if (c == 'c' && !strncmp (argv[0], "-command", length) && length > 3)
      command = argv[1];
    else if (c == 'c' && !strncmp (argv[0], "-count", length) && length > 3) {
      if (Tcl_GetInt (interp, argv[1], &massCnt) == TCL_ERROR)
	return TCL_ERROR;
      if (massCnt < 1) {
	sprintf (interp->result, "invalid -count %d", massCnt);
	return TCL_ERROR;
      }
    } else if (c == 'p' && !strncmp (argv[0], "-position", length)) {
      c      = argv[1][0];
      length = strlen (argv[1]);
      if (c == 'f' && !strncmp (argv[1], "first", length))
	target = "FIRST";
      else if (c == 'l' && !strncmp (argv[1], "last",  length))
	target = "LAST";
      else if (c == 'n' && !strncmp (argv[1], "next",  length))
	target = "NEXT";
      else if (c == 'p' && !strncmp (argv[1], "prev",  length))
	target = "PREV";
      else if (Tcl_GetInt (interp, argv[1], &pos) == TCL_OK) {
	sprintf (buf, "POS (%d)", pos);
	target = buf;
      } else {
	Tcl_AppendResult (interp, "bad value for switch -position \"", argv[1],
			  "\": must be -first, -last, -prev, -next or a ",
			  "number", (char *) NULL);
	return TCL_ERROR;
      }
    } else {
      Tcl_AppendResult (interp, "unknown or ambigious switch \"", argv[0],
			"\": must be -command, -count or -position",
			(char *) NULL);
      return TCL_ERROR;
    }
  }

  if (argc > 0) {
    Tcl_AppendResult (interp, "missing argument for switch: \"", argv[0],
		      "\"", (char *) NULL);
    return TCL_ERROR;
  }

  if (*target && massCnt > 0) {
    sprintf (interp->result, "Invalid combination of -count and -position");
    return TCL_ERROR;
  }
  if (*target && command) {
    sprintf (interp->result, "Invalid combination of -command and -position");
    return TCL_ERROR;
  }

  if (command) {
    if (massCnt < 1) massCnt = 20;
    return AdaFetchAll (interp, cursorName, massCnt, command);
  }
  if (!massCnt) massCnt = 1;

  /*
   * Send the fetch command to the database kernel.
   */
  sprintf (cmdBuf, "FETCH %s %s INTO ?", target, cursorVars->resultTable);
  if (AdabasSendCmdPacket (interp, cursorVars->logonInfo,
			   0, massCnt, cmdBuf) == TCL_ERROR) {
    if (strcmp (Tcl_GetVar2 (interp, ADABAS_MESSAGES, "rc", TCL_GLOBAL_ONLY),
		"100")) return TCL_ERROR;
    Tcl_ResetResult (interp);
    return TCL_OK;
  }

  if (!(data = getDataOfCursor (interp, cursorVars,
				massCnt > 1 ? MASS_CMD : 0)))
    return TCL_ERROR;

  nullValue = getNullValue (interp);
  ret       = TCL_OK;
  while (data && ret == TCL_OK) {
    if (massCnt > 1) {
      Tcl_AppendResult (interp, "{", (char *) NULL);
    }
    currSi  = cursorVars->shortInfos;
    for (ix = 0; ix < cursorVars->cntParams && ret == TCL_OK; ix++) {
      ret = appendOneValue (interp, data, nullValue, currSi++,
			    &cursorVars->longColFound, 1);
    }
    if (massCnt > 1) {
      Tcl_AppendResult (interp, "} ", (char *) NULL);
      data = getNextData (interp, data, cursorVars);
    } else {
      data = NULL;
    }
  }

  return ret;
}

/* -------------------------------------------------------------- */

static int appendOneValue (interp, data, nullValue, currSi,
			   longColFound, isElement)
     Tcl_Interp *interp;		/* Current interpreter. */
     char       *data;			/* Value to append as string. */
     char       *nullValue;		/* Representation of a NULL value. */
     field_info *currSi;		/* Shortfield info of value. */
     int        *longColFound;		/* Set to 1, if value is LONG. */
     int         isElement;		/* append result as element? */
{
  int ret;

#if ADABAS_VERSION >= 62
  if (currSi->io_type == IO_INPUT)
    return TCL_OK; /* nothing to append. */
#endif    

#if ADABAS_VERSION >= 62
  ret = AdabasAppendData (interp, data+currSi->bufpos-1,
			  currSi->data_type, currSi->in_out_len,
			  currSi->length, currSi->frac,
			  nullValue, isElement);
#else /* ADABAS_VERSION <= 61 */
  if (currSi->length) {
    ret = AdabasAppendData (interp, data+currSi->var.s.bufpos-1,
			    currSi->d_type,
			    (unsigned char) currSi->var.s.in_out_l,
			    (unsigned char) currSi->length,
			    currSi->var.s.frac, nullValue, isElement);
  } else {
    ret = AdabasAppendData (interp, data+currSi->var.s.bufpos-1,
			    currSi->d_type, currSi->var.l.longiolen,
			    currSi->var.l.longiolen-1, 0,
			    nullValue, isElement);
  }
#endif    

#if ADABAS_VERSION >= 62
  if (currSi->data_type == DSTRA  ||
      currSi->data_type == DSTRE  ||
      currSi->data_type == DSTRB  ||
      currSi->data_type == DSTRDB ||
      currSi->data_type == DSTRUNI)
    *longColFound = 1;
#else /* ADABAS_VERSION <= 61 */
  if (currSi->d_type == CSP_INFO_C_OLDLONG_CHAR  ||
      currSi->d_type == CSP_INFO_B_OLDLONG_BYTE  ||
      currSi->d_type == CSP_INFO_A_OLDLONG_ASCII_DBYTE)
    *longColFound = 1;
#endif    

  return ret;
}

/* -------------------------------------------------------------- */

static int AdaFetchAll (interp, cursorName, count, tclCommand)
     Tcl_Interp *interp;		/* Current interpreter. */
     char       *cursorName;		/* Name of current cursor. */
     int         count;			/* Count for mass fetches. */
     char       *tclCommand;		/* Command to execute for each row. */
{
  char *argv[5];
  char *command;
  int ret;
  char countBuf[30];

  sprintf (countBuf, "%d", count);

  argv[0] = "_adafetch_all";
  argv[1] = cursorName;
  argv[2] = countBuf;
  argv[3] = tclCommand;
  argv[4] = (char *) NULL;

  command = Tcl_Merge (4, argv);
  ret     = Tcl_Eval (interp, command);
  ckfree (command);
  return ret;
}

/* -------------------------------------------------------------- */

static char *getNextData (interp, data, cursorVars)
     Tcl_Interp *interp;		/* Current interpreter. */
     char       *data;			/* Buffer of data part. */
     CursorInfo *cursorVars;		/* Current cursor. */
{
  field_info *lastSi;
  int4        nextOff;
#if ADABAS_VERSION <= 61
  int2        lastIOLength;
#endif

  cursorVars->rowIx++;
  lastSi  = cursorVars->shortInfos+(cursorVars->cntParams-1);
#if ADABAS_VERSION >= 62
  nextOff = lastSi->bufpos+lastSi->in_out_len-1;
#else /* ADABAS_VERSION <= 61 */
  if (lastSi->length) {
    lastIOLength = (unsigned char) lastSi->var.s.in_out_l;
  } else {
    lastIOLength = lastSi->var.l.longiolen;
  }
  nextOff = lastSi->var.s.bufpos+lastIOLength-1;
#endif

  if (cursorVars->rowIx >= cursorVars->massRows ||
      cursorVars->dataEnd + nextOff >= cursorVars->dataLength) {
    /*
     * We reached the end of data, so better return a NULL.
     */
    return NULL;
  }

  /*
   * Increase the dataEnd marker in cursorVars and
   * return the adjusted data pointer.
   */
  cursorVars->dataEnd += nextOff;
  return data+nextOff;
}

/* -------------------------------------------------------------- */

static int AdaColsCmd (dummy, interp, argc, argv)
     ClientData dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int argc;				/* Number of arguments. */
     char **argv;			/* Argument strings. */
{
  CursorInfo *cursorVars;
  char        name[31];
  char        number[20];
  int         ix;
  char       *dataType;
  int         length;
  int         frac;
  field_info *currSi;
#if ADABAS_VERSION >= 62
  char       *currCol;
#endif

  if (argc != 2) {
    Tcl_AppendResult (interp, "wrong # args: should be \"", argv[0],
		      " cursorHandle\"", (char *) NULL);
    return TCL_ERROR;
  }

  if (!getCursorHandle (interp, argv[1], &cursorVars)) return TCL_ERROR;

  currSi  = cursorVars->shortInfos;
#if ADABAS_VERSION >= 62
  currCol = cursorVars->columnNames;
#endif

  Tcl_UnsetVar2 (interp, ADABAS_MESSAGES, "coltypes",   TCL_GLOBAL_ONLY);
  Tcl_UnsetVar2 (interp, ADABAS_MESSAGES, "collengths", TCL_GLOBAL_ONLY);
  Tcl_UnsetVar2 (interp, ADABAS_MESSAGES, "colprecs",   TCL_GLOBAL_ONLY);
  Tcl_UnsetVar2 (interp, ADABAS_MESSAGES, "colscales",  TCL_GLOBAL_ONLY);

  for (ix = 0; ix < cursorVars->cntParams; ix++) {
#if ADABAS_VERSION >= 62
    dataType = dataType2String (currSi->data_type);
    length   = currSi->length;
    frac     = currSi->frac;
    sprintf (name, "%.*s", *currCol, currCol+1);
    currCol += *currCol + 1;
#else /* ADABAS_VERSION <= 61 */
    dataType = dataType2String (currSi->d_type);
    if (currSi->length) {
      length = (unsigned char) currSi->length;
      frac   = currSi->var.s.frac;
    } else {
      length = currSi->var.l.longiolen - 1;
      frac   = 0;
    }
    p2c (name, currSi->fieldname, sizeof (currSi->fieldname));
#endif
    currSi++;
    Tcl_SetVar2 (interp, ADABAS_MESSAGES, "coltypes",   dataType,
		 TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);
    sprintf (number, "%d", length);
    Tcl_SetVar2 (interp, ADABAS_MESSAGES, "collengths", number,
		 TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);
    Tcl_SetVar2 (interp, ADABAS_MESSAGES, "colprecs",   number,
		 TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);
    sprintf (number, "%d", frac);
    Tcl_SetVar2 (interp, ADABAS_MESSAGES, "colscales",  number,
		 TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);

    Tcl_AppendElement (interp, name);
  }

  return TCL_OK;
}

/* -------------------------------------------------------------- */

static int AdaReadLongCmd (dummy, interp, argc, argv)
     ClientData dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int argc;				/* Number of arguments. */
     char **argv;			/* Argument strings. */
{
  CursorInfo     *cursorVars;
  long_descriptor longDesc;
  int             length;
  int             c;
  char           *cursorName;
  char           *tableName   = (char *) NULL;
  char           *columnName  = (char *) NULL;
  char           *descrString = (char *) NULL;
  char           *fileName    = (char *) NULL;
  char           *whereCond   = (char *) NULL;
  long            bytes;
  int             cont;
  char           *nullValue;

  if (argc <= 2) {
    Tcl_AppendResult (interp, "wrong # args: should be \"", argv[0],
		      " cursorHandle options...\"", (char *) NULL);
    return TCL_ERROR;
  }
  cursorName = argv[1];

  for (argv+=2, argc-=2; argc > 1; argv += 2, argc -= 2) {
    c        = argv[0][1];
    length   = strlen (argv[0]);

    if (c == 'c' && !strncmp (argv[0], "-column", length)) {
      columnName = argv[1];
    } else if (c == 'd' && !strncmp (argv[0], "-descriptor", length)) {
      descrString = argv[1];
    } else if (c == 'f' && !strncmp (argv[0], "-filename", length)) {
      fileName = argv[1];
    } else if (c == 't' && !strncmp (argv[0], "-table", length)) {
      tableName = argv[1];
    } else if (c == 'w' && !strncmp (argv[0], "-where", length)) {
      whereCond = argv[1];
    } else {
      Tcl_AppendResult (interp, "bad switch \"", argv[0], "\": must be ",
			"-table, -column, -where, -descriptor or -filename",
			(char *) NULL);
      return TCL_ERROR;
    }
  }
  if (argc > 0) {
    Tcl_AppendResult (interp, "missing argument for switch: \"", argv[0],
		      "\"", (char *) NULL);
    return TCL_ERROR;
  }
  if (!(( descrString && !tableName && !columnName && !whereCond) ||
	(!descrString &&  tableName &&  columnName &&  whereCond))) {
    interp->result = "invalid combination of switches.";
    return TCL_ERROR;
  }

  if (!getCursorHandle (interp, cursorName, &cursorVars)) return TCL_ERROR;

  if (descrString) {
    /*
     * The variant with the long descriptor is given.
     * First check, if it is a nullvalue.
     */
    nullValue = getNullValue (interp);
    if (!strcmp (descrString, nullValue)) {
      Tcl_SetResult (interp, nullValue, TCL_STATIC);
      return TCL_OK;
    }

    if (list2LongDescr (interp, descrString, &longDesc) == TCL_ERROR)
      return TCL_ERROR;

    if (!cursorVars->longColFound) {
      interp->result = "call of adareadlong must be preceded by adafetch";
      return TCL_ERROR;
    }

  } else {
    /*
     * The variant with table, column and condition is given.
     * So first get a new copy of the cursor handle, then start the SELECT.
     */
    if (!(cursorVars = allocCursor (interp, cursorVars->logonInfo)))
      return TCL_ERROR;

    if (AdabasLongSelect (interp, cursorVars, tableName, columnName,
			  whereCond, &longDesc) == TCL_ERROR) {
      freeCursor (cursorVars);
      return TCL_ERROR;
    }
  }

  if (longDesc.valmode == VM_NODATA) {
    bytes = 0;
    cont  = 1;
  } else {
    if (AdabasLongReadFirst (interp, cursorVars, &longDesc,
			     fileName, &bytes, &cont) == TCL_ERROR) {
      if (!descrString) freeCursor (cursorVars);
      return TCL_ERROR;
    }
  }

  while (cont) {
    if (AdabasLongReadNext (interp, cursorVars, &longDesc,
			    fileName, &bytes, &cont) == TCL_ERROR) {
      if (!descrString) freeCursor (cursorVars);
      return TCL_ERROR;
    }
  }

  /*
   * If called with a filename to write in, the number of
   * bytes written are returned.
   */
  if (fileName)
    sprintf (interp->result, "%ld", bytes);

  if (!descrString) freeCursor (cursorVars);
  return TCL_OK;
}

/* -------------------------------------------------------------- */

static int AdaWriteLongCmd (dummy, interp, argc, argv)
     ClientData dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int argc;				/* Number of arguments. */
     char **argv;			/* Argument strings. */
{
  CursorInfo     *cursorVars;
  long_descriptor longDesc;
  int             length;
  int             c;
  int4            longLength;
  int4            bytes;
  int             cont;
  Tcl_DString     parsId;
  Tcl_DString     substVal;
  char           *cursorName;
  char           *tableName  = (char *) NULL;
  char           *columnName = (char *) NULL;
  char           *fileName   = (char *) NULL;
  char           *longValue  = (char *) NULL;
  char           *whereCond  = (char *) NULL;

  if (argc <= 2) {
    Tcl_AppendResult (interp, "wrong # args: should be \"", argv[0],
		      " cursorHandle options...\"", (char *) NULL);
    return TCL_ERROR;
  }
  cursorName = argv[1];

  for (argv+=2, argc-=2; argc > 1; argv += 2, argc -= 2) {
    c        = argv[0][1];
    length   = strlen (argv[0]);

    if (c == 'c' && !strncmp (argv[0], "-column", length)) {
      columnName = argv[1];
    } else if (c == 'f' && !strncmp (argv[0], "-filename", length)) {
      fileName = argv[1];
    } else if (c == 't' && !strncmp (argv[0], "-table", length)) {
      tableName = argv[1];
    } else if (c == 'v' && !strncmp (argv[0], "-value", length)) {
      longValue = argv[1];
    } else if (c == 'w' && !strncmp (argv[0], "-where", length)) {
      whereCond = argv[1];
    } else {
      Tcl_AppendResult (interp, "bad switch \"", argv[0], "\": must be ",
			"-table, -column, -where, -value or -filename",
			(char *) NULL);
      return TCL_ERROR;
    }
  }
  if (argc > 0) {
    Tcl_AppendResult (interp, "missing argument for switch: \"", argv[0],
		      "\"", (char *) NULL);
    return TCL_ERROR;
  }
  if (!tableName && !columnName && !whereCond) {
    interp->result = "invalid combination of switches.";
    return TCL_ERROR;
  }
  if (!longValue && !fileName) {
    interp->result = "either -filename or -value switch must be given";
    return TCL_ERROR;
  }

  if (!getCursorHandle (interp, cursorName, &cursorVars))
    return TCL_ERROR;
  if (!(cursorVars = allocCursor (interp, cursorVars->logonInfo)))
    return TCL_ERROR;

  /*
   * Send the constructed UPDATE sql command to the kernel for parsing.
   */
  if (AdabasLongUpdate (interp, cursorVars, tableName, columnName,
			whereCond, &parsId) == TCL_ERROR) {
    freeCursor (cursorVars);
    return TCL_ERROR;
  }

  if (AdabasLongValLength (interp, fileName, longValue,
			   &substVal, &longLength) == TCL_ERROR) {
    freeCursor (cursorVars);
    return TCL_ERROR;
  }
    
  /*
   * Now the parsid will be executed, the first part of the long value
   * (or all, if it fits into the packet) is given as data.
   */
  if (AdabasLongWriteFirst (interp, cursorVars, &longDesc, &parsId, fileName,
			    Tcl_DStringValue (&substVal), longLength,
			    &bytes, &cont) == TCL_ERROR) {
    Tcl_DStringFree (&parsId);
    if (longValue) Tcl_DStringFree (&substVal);
    freeCursor (cursorVars);
    return TCL_ERROR;
  }

  if (cont) {
    /*
     * While not all of the long value is written,
     * continue with sending PUTVAL packets.
     */
    while (cont) {
      if (AdabasLongWriteNext (interp, cursorVars, &longDesc, fileName,
			       Tcl_DStringValue (&substVal), longLength,
			       &bytes, &cont) == TCL_ERROR) {
	Tcl_DStringFree (&parsId);
	if (longValue) Tcl_DStringFree (&substVal);
	freeCursor (cursorVars);
	return TCL_ERROR;
      }
    }

    /*
     * Finally inform the kernel, that everything is okay and over.
     * ignore the return value, since we are in the clean up phase now.
     (void) AdabasLongWriteEnd (interp, cursorVars, &longDesc);
     * we better comment this out...
     */
  }

  /*
   * Finally throw away the generated parsid; ignore its return value, too.
   */
  (void) AdabasSend2PartPacket (interp, cursorVars->logonInfo,
				pkCommand, "DROP PARSEID", 0,
				pkParsid,  Tcl_DStringValue (&parsId), 0);

  Tcl_DStringFree (&parsId);
  if (longValue) Tcl_DStringFree (&substVal);
  freeCursor (cursorVars);
  return TCL_OK;
}

/* -------------------------------------------------------------- */

static int AdaAutoCommitCmd (dummy, interp, argc, argv)
     ClientData dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int argc;				/* Number of arguments. */
     char **argv;			/* Argument strings. */
{
  AdabasInfo *vars;
  int         onOff;

  if (argc != 3) {
    Tcl_AppendResult (interp, "wrong # args: should be \"", argv[0],
		      " logonHandle onOff\"", (char *) NULL);
    return TCL_ERROR;
  }

  if (getLogonHandle (interp, argv[1], &vars) == TCL_ERROR)
    return TCL_ERROR;

  if (Tcl_GetBoolean (interp, argv[2], &onOff) == TCL_ERROR)
    return TCL_ERROR;

  vars->packetInfo.commitImmediately = onOff;

  return TCL_OK;
}

/* -------------------------------------------------------------- */

static int AdaCommitCmd (dummy, interp, argc, argv)
     ClientData dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int argc;				/* Number of arguments. */
     char **argv;			/* Argument strings. */
{
  AdabasInfo *vars;

  if (argc != 2) {
    Tcl_AppendResult (interp, "wrong # args: should be \"", argv[0],
		      " logonHandle\"", (char *) NULL);
    return TCL_ERROR;
  }

  if (getLogonHandle (interp, argv[1], &vars) == TCL_ERROR)
    return TCL_ERROR;

  return AdabasSendCmdPacket (interp, vars, 0, 1, "COMMIT WORK");
}

/* -------------------------------------------------------------- */

static int AdaRollbackCmd (dummy, interp, argc, argv)
     ClientData dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int argc;				/* Number of arguments. */
     char **argv;			/* Argument strings. */
{
  AdabasInfo *vars;

  if (argc != 2) {
    Tcl_AppendResult (interp, "wrong # args: should be \"", argv[0],
		      " logonHandle\"", (char *) NULL);
    return TCL_ERROR;
  }

  if (getLogonHandle (interp, argv[1], &vars) == TCL_ERROR)
    return TCL_ERROR;

  return AdabasSendCmdPacket (interp, vars, 0, 1, "ROLLBACK WORK");
}

/* -------------------------------------------------------------- */

static int AdaUtilCmd (dummy, interp, argc, argv)
     ClientData dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int argc;				/* Number of arguments. */
     char **argv;			/* Argument strings. */
{
  AdabasInfo *vars;
  char       *data;
  int4        dataLength;
  int         argCount;
  int         cntParams;
  int         ret;
  int         ix;
  int         longColFound;
  char       *nullValue;
  char        dbName[200]; /* more than enough. */  
  Tcl_DString resultStr;
  field_info *currSi;
  field_info *shortInfos = NULL;

  if (argc != 3) {
    Tcl_AppendResult (interp, "wrong # args: should be \"", argv[0],
		      " logonHandle command\"", (char *) NULL);
    return TCL_ERROR;
  }

  if (getLogonHandle (interp, argv[1], &vars) == TCL_ERROR)
    return TCL_ERROR;

  if (vars->packetInfo.messType != M_UTILITY) {
    Tcl_AppendResult (interp, "Logon handle \"", argv[1],
		      "\" must be utility service", (char *) NULL);
    return TCL_ERROR;
  }

  if (AdabasSendCmdPacket (interp, vars, 0, 1, argv[2]) == TCL_ERROR)
    return TCL_ERROR;

#if ADABAS_VERSION >= 62
  data = getData (&vars->rcvPacket->segm, &dataLength, &argCount, UTILITY_CMD);
#else /* ADABAS_VERSION <= 61 */
  data = getData (&vars->sqlPacket, &dataLength, &argCount, UTILITY_CMD);
#endif
  if (data && dataLength) {
#if ADABAS_VERSION >= 62
    cntParams = getShortInfos (&vars->rcvPacket->segm, &shortInfos);
#else /* ADABAS_VERSION <= 61 */
    cntParams = 0;
#endif
    if (cntParams && shortInfos) {
      nullValue = getNullValue (interp);
      currSi    = shortInfos;
      ret       = TCL_OK;
      for (ix = 0; ix < cntParams && ret == TCL_OK; ix++) {
	ret = appendOneValue (interp, data, nullValue, currSi++,
			      &longColFound, 1);
      }

    } else 
      switch (*data) {
      case  2:
	p2c (dbName, data+1, dataLength-1);
	Tcl_AppendElement (interp, "warm");
	Tcl_AppendElement (interp, dbName);
	break;
      case 0: case 1:
	Tcl_AppendElement (interp, "cold");
	break;
      default:
	Tcl_DStringInit   (&resultStr);
	Tcl_DStringAppend (&resultStr, data, dataLength);
	Tcl_DStringResult (interp, &resultStr);
    }
    if (shortInfos) ckfree ((char *) shortInfos);
  }
  return TCL_OK;
}

/* -------------------------------------------------------------- */

static int AdaLogonCmd (dummy, interp, argc, argv)
     ClientData dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int argc;				/* Number of arguments. */
     char **argv;			/* Argument strings. */
{
  int          length;
  int          c;
  Tcl_RegExp   regExp;
  Tcl_DString  connect;
  Tcl_DString  data;
  Tcl_DString  cUser;
  Tcl_DString  cCryptPasswd;
  Tcl_DString  cDbName;
  Tcl_DString  help;
  char         xUser[19];
  char         xCryptPasswd[24];
  char         xNode[65];
  char         xDbName[19];
  char        *user;
  AdabasInfo  *vars;
  char        *xuserKey;
  char        *start;
  char        *end;
  int          ret;
  char        *connectStr;
  char        *sqlMode     = (char *) NULL;
  service      service     = SQL_USER;
  int          controlAppl = 0;
  xuser_record userParams;

  if (argc < 2) {
    Tcl_AppendResult (interp, "wrong # args: should be \"", argv[0],
		      " connectStr ?options?\"", (char *) NULL);
    return TCL_ERROR;
  }
  connectStr = argv[1];

  for (argv+=2, argc-=2; argc > 1; argv += 2, argc -= 2) {
    c        = argv[0][1];
    length   = strlen (argv[0]);

    if (c == 's' && !strncmp (argv[0], "-service", length) && length > 1) {
      if (!strcmp (argv[1], "control")) {
	controlAppl = 1;
      } else if ((service = scanService (interp, argv[1])) < 0) {
	Tcl_AppendResult (interp, "Unknown service \"", argv[1],
			  "\": must be user, utility or control",
			  (char *) NULL);
	return TCL_ERROR;
      }
    } else if (c == 's' && !strncmp (argv[0], "-sqlmode", length)
	       && length > 1) {
      if (scanSqlMode (interp, argv[1], 1) < 0)
	return TCL_ERROR;
      sqlMode = argv[1];
    } else {
      Tcl_AppendResult (interp, "bad switch \"", argv[0], "\": must be ",
			"-service or -sqlmode", (char *) NULL);
      return TCL_ERROR;
    }
  }
  if (argc > 0) {
    Tcl_AppendResult (interp, "missing argument for switch: \"", argv[0],
		      "\"", (char *) NULL);
    return TCL_ERROR;
  }
  if (service == SQL_UTILITY && sqlMode) {
    interp->result = "invalid combination of switches.";
    return TCL_ERROR;
  }

  regExp = Tcl_RegExpCompile (interp, "^([^,]*),(.*)$");
  if (!Tcl_RegExpExec (interp, regExp, connectStr, connectStr)) {
    Tcl_AppendResult (interp, "Invalid connectString \"", connectStr,
		      "\", must be user,passwd", (char *) NULL);
    return TCL_ERROR;
  }
  Tcl_DStringInit (&help);
  Tcl_DStringInit (&cUser);
  Tcl_DStringInit (&cCryptPasswd);

  Tcl_RegExpRange (regExp, USER_RANGE, &start, &end);
  if (start == end) {
    /*
     * There was no user given; perhaps the name behind
     * the comma is meant as an xuser key.
     */
    Tcl_RegExpRange (regExp, PASSWD_RANGE, &start, &end);
    if (start == end) {
      /*
       * There was no xuser key given either.
       * So let's look after the default user.
       */
      xuserKey = NULL;
    } else {
      Tcl_DStringAppend (&help, start, end-start);
      xuserKey = Tcl_DStringValue (&help);
    }
    if (AdabasXUserGet (interp, xuserKey, &userParams) == TCL_ERROR) {
      Tcl_DStringFree (&help);
      return TCL_ERROR;
    }

    p2c (xUser, userParams.xu_user,
	 sizeof (userParams.xu_user));
    p2c (xDbName, userParams.xu_serverdb,
	 sizeof (userParams.xu_serverdb));
    p2c (xNode, userParams.xu_servernode,
	 sizeof (userParams.xu_servernode));
    memcpy (xCryptPasswd, userParams.xu_password,
	    sizeof (userParams.xu_password));

    Tcl_DStringFree (&help);
    escapeSequence (&cCryptPasswd, xCryptPasswd, sizeof (xCryptPasswd));
    user = xUser;
    if (*xDbName && *xDbName != ' ') {
      Tcl_DStringAppend (&help,  "set env(DBNAME) ", -1);
      if (*xNode && *xNode != ' ') {
	Tcl_DStringAppend (&help, xNode, -1);
	Tcl_DStringAppend (&help, ":",    1);
      }
      Tcl_DStringAppend (&help,  xDbName, -1);
      (void) Tcl_Eval (interp, Tcl_DStringValue (&help));
      Tcl_DStringFree   (&help);
    }
  } else {
    /*
     * There was really a pair of user and password given.
     */
    Tcl_DStringAppend (&cUser, start, end-start); 
    user = Tcl_DStringValue (&cUser);

    Tcl_RegExpRange   (regExp, PASSWD_RANGE, &start, &end);
    Tcl_DStringAppend (&help, start, end-start);

    /*
     * Change passwd into uppercase before crypting it.
     */
    for (start = Tcl_DStringValue (&help); *start; start++)
      *start = toupper (*start);

    AdabasPasswdCrypt (&cCryptPasswd, Tcl_DStringValue (&help));
    Tcl_DStringFree   (&help);
  }

  switch (service) {
  case  SQL_UTILITY:
    if (Tcl_Eval (interp, "adabas open -service utility") == TCL_ERROR)
      return TCL_ERROR;
    break;
  case SQL_USER:
    if (Tcl_Eval (interp, "adabas open") == TCL_ERROR)
      return TCL_ERROR;
    break;
  }

  Tcl_DStringInit (&cDbName);
  Tcl_DStringAppend (&cDbName, interp->result, -1);

  if (getLogonHandle (interp, Tcl_DStringValue (&cDbName),
		      &vars) == TCL_ERROR) {
    (void) AdabasCloseCmd (interp, Tcl_DStringValue (&cDbName), vars);
    Tcl_DStringFree (&cDbName);
    return TCL_ERROR;
  }

  if (controlAppl) {
    memcpy (vars->packetInfo.application, "CON", 3);
  } else if (service == SQL_UTILITY) {
    vars->packetInfo.messType = M_UTILITY;
  }

  Tcl_DStringInit   (&connect);
  Tcl_DStringAppend (&connect, "CONNECT ", -1);
  Tcl_DStringAppend (&connect, user,       -1);
  Tcl_DStringAppend (&connect, " IDENTIFIED BY :x", -1);

  if (sqlMode) {
    Tcl_DStringAppend (&connect, " SQLMODE ", -1);
    Tcl_DStringAppend (&connect,   sqlMode,   -1);
  }

  Tcl_DStringInit   (&data);
  escapeSequence    (&data, "", 1); /* defined byte */
  Tcl_DStringAppend (&data, Tcl_DStringValue (&cCryptPasswd), -1);
  Tcl_DStringAppend (&data, " ", 1);
  AdabasRteTermId   (Tcl_DStringValue (&help)); 
  Tcl_DStringAppend (&data, Tcl_DStringValue (&help), -1); 

  ret = AdabasSend2PartPacket (interp, vars,
			       pkCommand, Tcl_DStringValue (&connect), 0,
			       pkData,    Tcl_DStringValue (&data),    0);
  if (ret)
    (void) AdabasCloseCmd (interp, Tcl_DStringValue (&cDbName), vars);
  
  Tcl_DStringFree (&connect);
  Tcl_DStringFree (&data);
  Tcl_DStringFree (&help);
  Tcl_DStringFree (&cDbName);

  return ret;
}

/* -------------------------------------------------------------- */

static int AdaLogoffCmd (dummy, interp, argc, argv)
     ClientData dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int argc;				/* Number of arguments. */
     char **argv;			/* Argument strings. */
{
  AdabasInfo  *vars;

  if (argc != 2) {
    Tcl_AppendResult (interp, "wrong # args: should be \"", argv[0],
		      " logonHandle\"", (char *) NULL);
    return TCL_ERROR;
  }

  if (getLogonHandle (interp, argv[1], &vars) == TCL_ERROR)
    return TCL_ERROR;

  (void) AdabasSendCmdPacket (interp, vars, 0, 1, "ROLLBACK WORK RELEASE");
  (void) AdabasCloseCmd (interp, argv[1], vars);
  return TCL_OK;
}

/* -------------------------------------------------------------- */

static Tcl_HashEntry *getCursorHandle (interp, cursorName, cursorVars)
     Tcl_Interp  *interp;		/* Current interpreter. */
     char        *cursorName;		/* Name to look for. */
     CursorInfo **cursorVars;		/* Returned cursor. */
{
  Tcl_HashEntry *hashEntry;

  if (!(hashEntry = Tcl_FindHashEntry (&cursorHash, cursorName))) {
    Tcl_AppendResult (interp, "No valid cursor handle: ",
		      cursorName, (char *) NULL);
    return NULL;
  }
  *cursorVars = (CursorInfo *) Tcl_GetHashValue (hashEntry);
  Tcl_SetVar2 (interp, ADABAS_MESSAGES, "handle", cursorName, TCL_GLOBAL_ONLY);
  return hashEntry;
}

/* -------------------------------------------------------------- */

static int getLogonHandle (interp, logonName, logon)
     Tcl_Interp  *interp;		/* Current interpreter. */
     char        *logonName;		/* Name to look for. */
     AdabasInfo **logon;		/* Returned Login. */
{
  Tcl_CmdInfo cmdInfo;

  if (!Tcl_GetCommandInfo (interp, logonName, &cmdInfo)
      || cmdInfo.proc != AdabasSendCmd) {
    Tcl_AppendResult (interp, "Invalid logonHandle \"",
		      logonName, "\"", (char *) NULL);
    return TCL_ERROR;
  }
  *logon = (AdabasInfo *) cmdInfo.clientData;
  return TCL_OK;
}

/* -------------------------------------------------------------- */

static void examineSqlCommand (interp, cmdString, withInfo, withResCount,
			       selectKind, selectInto)
     Tcl_Interp *interp;		/* Current interpreter. */
     char       *cmdString;		/* String containing a SQL command. */
     int        *withInfo;		/* SQL, that returns results? */
     int        *withResCount;		/* SQL, that returns rowcnt? */
     SelectKind *selectKind;		/* returned kind of sql statement. */
     char      **selectInto;		/* position of first parameter name
					 * in cmdString (for SELECT INTO). */
{
  Tcl_DString cmd;
  Tcl_RegExp  regExp;
  char       *start;
  char       *end;

  *withInfo     = 0;
  *withResCount = 0;
  *selectKind   = noSelect;
  *selectInto   = (char *) NULL;

  regExp = Tcl_RegExpCompile (interp, "^[ \t\n]*([^ \t\n(]*)");
  if (!Tcl_RegExpExec (interp, regExp, cmdString, cmdString)) return;

  Tcl_RegExpRange   (regExp, CMD_RANGE, &start, &end);
  Tcl_DStringInit   (&cmd);
  Tcl_DStringAppend (&cmd, start, end-start);

  for (start = Tcl_DStringValue (&cmd); *start; start++)
    *start = toupper (*start);

  if (!strcmp (Tcl_DStringValue (&cmd), "SELECT")) {
    *selectKind = massSelect;
    *withInfo   = 1;
    /*
     * Now that we found a SELECT, lets look if it is a SELECT INTO.
     */
    Tcl_DStringFree   (&cmd);
    Tcl_DStringAppend (&cmd, cmdString, -1);
    for (start = Tcl_DStringValue (&cmd); *start; start++)
      *start = toupper (*start);
    regExp = Tcl_RegExpCompile (interp, " INTO  *(:)");
    if (Tcl_RegExpExec (interp, regExp,
			Tcl_DStringValue (&cmd), Tcl_DStringValue (&cmd))) {
      Tcl_RegExpRange (regExp, 1, &start, &end);
      *selectKind = intoSelect;
      *selectInto = cmdString + (start - Tcl_DStringValue (&cmd));
    }

  } else if (!strcmp (Tcl_DStringValue (&cmd), "EXPLAIN")) {
    *withInfo     = 1;
    *selectKind   = explainSelect;

  } else if (!strcmp (Tcl_DStringValue (&cmd), "SHOW")) {
    *withInfo     = 1;
    *selectKind   = showSelect;

  } else if (!strcmp (Tcl_DStringValue (&cmd), "UPDATE") ||
	     !strcmp (Tcl_DStringValue (&cmd), "INSERT") ||
	     !strcmp (Tcl_DStringValue (&cmd), "DELETE")) {
    *withResCount = 1;
  }

  Tcl_DStringFree (&cmd);
}

/* -------------------------------------------------------------- */

static char *getNullValue (interp)
     Tcl_Interp *interp;		/* Current interpreter. */
{
  char *nullValue = Tcl_GetVar2 (interp, ADABAS_MESSAGES, "nullvalue",
				 TCL_GLOBAL_ONLY);
  return nullValue ? nullValue : "";
}
