/* -------------------------------------------------------------- */
/*  LAST EDIT: Sat Aug 17 18:46:41 1996 by Adabas D Database (gutamil.v12-berlin.de!adabas)  */
/*  $Id: AdabasTcl.c,v 1.64 1996/08/17 16:46:48 adabas Exp $ */
/* -------------------------------------------------------------- */
/* Copyright (c) 1996 Christian Krone. All rights reserved.       */
/* This program is free software; you can redistribute it and/or  */
/* modify it under the same terms as Tcl itself.                  */
/* See also licence.terms                                         */
/* -------------------------------------------------------------- */

/* ----------------- Import interface --------------------------- */
#include <stdlib.h>
#include <string.h>
#include <tcl.h>
#ifdef ADABAS_WISH
#include <tk.h>
#endif

#include "adabas.h"
#include "sqlrte.h"

#include "AdabasUtil.h"
#include "AdabasFormat.h"
#include "AdabasRte.h"
#include "AdabasPacket.h"
#include "AdabasSend.h"
#include "AdabasSql.h"
#include "AdabasLong.h"

/* ----------------- Type definitions --------------------------- */

typedef struct {
  char *name;
  enum {oWithInfo=1, oMessType, oProducer, oSqlMode,
	oApplication, oMassCmd, oCommitImmediately} kind;
} OptionMap;

/* ----------------- Exported procedure headings ---------------- */
#ifdef ADABAS_MODULE
extern int Adabastcl_Init _ANSI_ARGS_((Tcl_Interp *interp));
#else
extern int main _ANSI_ARGS_((int argc, char **argv));
#endif

/* ----------------- Local variables ---------------------------- */
static OptionMap sendOptions[] = {
  { "-withInfo",          oWithInfo          },
  { "-messType",          oMessType          },
  { "-producer",          oProducer          },
  { "-sqlMode",           oSqlMode           },
  { "-application",       oApplication       },
  { "-massCmd",           oMassCmd           },
  { "-commitImmediately", oCommitImmediately },
  { (char *) NULL, 0 } /* termination flag */
};

static partKind partKinds[] = {
  pkCommand,
  pkParsid,
  pkResultcount,
  pkData,
  pkLongdata,
  pkNil /* termination flag */
};

/* ----------------- Local procedure headings ------------------- */
#ifndef ADABAS_MODULE
static int Adabastcl_Init _ANSI_ARGS_((Tcl_Interp *interp));
static int static_Init    _ANSI_ARGS_((Tcl_Interp *interp));
#endif
static int AdabasCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
				  int argc, char **argv));
static int AdabasOpenCmd _ANSI_ARGS_((ClientData dummy, Tcl_Interp *interp,
				      int argc, char **argv));
static int AdabasXGetCmd _ANSI_ARGS_((ClientData dummy, Tcl_Interp *interp,
				      int toFile, int argc, char **argv));
static int AdabasVersion _ANSI_ARGS_((ClientData dummy, Tcl_Interp *interp,
				      int argc, char **argv));
static int AdabasXUser _ANSI_ARGS_((ClientData dummy, Tcl_Interp *interp,
				    int argc, char **argv));
static int args2xuser _ANSI_ARGS_((Tcl_Interp *interp, int argc, char **argv,
				   xuser_record *userParams));
static void xuser2result _ANSI_ARGS_((Tcl_Interp *interp,
				      xuser_record *userParams,
				      char *givenPasswd));
static int AdabasSendConfigure _ANSI_ARGS_((Tcl_Interp *interp,
					    AdabasPacketInfo *vars,
					    char *optionName));
static int scanPacketConfOpts _ANSI_ARGS_((Tcl_Interp *interp, char *expected,
					   char *optionName,
					   int argc, char **argv,
					   AdabasPacketInfo *vars));

/*
 * All the procedures, that are used to fill a packet, have the packet
 * pointer as parameter and are therefore version dependend...
 */
#if ADABAS_VERSION >= 62
static void fillPacketHeader _ANSI_ARGS_((packet_header *packetHead,
					  int size, int swapKind,
					  char *application));
static void fillSegmentHeader _ANSI_ARGS_((packet_header *packetHead,
					   segment_header *segmentHead,
					   AdabasPacketInfo *options));
static int fillPart _ANSI_ARGS_((Tcl_Interp *interp, packet_header *packetHead,
				 segment_header *segmentHead, part **part,
				 int kind, AdabasPartInfo *partInfo));
static void setResultSegment _ANSI_ARGS_((Tcl_Interp *interp,
					  segment *retSegment));

#else /* ADABAS_VERSION <= 61 */
static void fillPacketHeader _ANSI_ARGS_((packet *packetHead, int swapKind,
					  char *application));
static void fillSegmentHeader _ANSI_ARGS_((packet *packetHead,
					   AdabasPacketInfo *options));
static int fillPart _ANSI_ARGS_((Tcl_Interp *interp, packet *packetHead,
				 partKind kind, AdabasPartInfo *partInfo,
				 int massCmd));
static void setResultSegment _ANSI_ARGS_((Tcl_Interp *interp,
					  packet *retSegment,
					  int sndMessType));
#endif

/* ----------------- Exported procedures ------------------------- */

#ifdef ADABAS_TCLSH
int main (argc, argv)
     int argc;				/* Number of arguments. */
     char **argv;			/* Argument strings. */
{
  Tcl_Main (argc, argv, Adabastcl_Init);
  return 0;
}
#endif

#ifdef ADABAS_WISH
int main (argc, argv)
     int argc;				/* Number of arguments. */
     char **argv;			/* Argument strings. */
{
  Tk_Main (argc, argv, Adabastcl_Init);
  return 0;
}
#endif

#ifndef ADABAS_MODULE
static
#endif
int Adabastcl_Init (interp)
     Tcl_Interp *interp;		/* Current interpreter. */
{
  char *adabasTclDir = getenv ("ADABASTCL_DIR");
  char  versionString[100];
  int   versionNumber;
  Tcl_RegExp regExp;
  char *start, *end;

#if defined ADABAS_TCLSH || defined ADABAS_WISH
  if (Tcl_Init(interp) == TCL_ERROR)
    return TCL_ERROR;
#endif

#ifdef ADABAS_WISH
  if (Tk_Init(interp) == TCL_ERROR)
    return TCL_ERROR;
#endif

  /*
   * Experience has pointed out, that it is very important to talk to
   * a database server of the correct version; so here we check, if
   * ADABAS_VERSION is configured correctly...
   */
  AdabasRteVersion (versionString);
  regExp = Tcl_RegExpCompile (interp, "RTE *(([0-9])\\.([0-9])[0-9.]*)");
  if (Tcl_RegExpExec (interp, regExp, versionString, versionString) <= 0) {
    versionNumber = 0;
  } else {
    Tcl_RegExpRange (regExp, 2, &start, &end);
    versionNumber  = (*start-'0')*10;
    Tcl_RegExpRange (regExp, 3, &start, &end);
    versionNumber += (*start-'0');
  }
  if ((ADABAS_VERSION <= 61 && versionNumber > 61) ||
      (ADABAS_VERSION >= 62 && versionNumber < 62)) {
#if ADABAS_VERSION <= 61
    char *wanted = "upto version 6.1.1";
#else /* ADABAS_VERSION >= 62 */
    char *wanted = "of version 6.2 or above";
#endif
    char found[80];

    if (versionNumber) {
      Tcl_RegExpRange (regExp, 1, &start, &end);
      sprintf (found, "%.*s", end-start, start);
    } else
      strcpy  (found, "unknown");
    Tcl_AppendResult (interp, "AdabasTcl was configured for a database ",
		      wanted, ", but found a server of version ",
		      found, (char *) NULL);
    return TCL_ERROR;
  }

  /*
   * Here now comes the core of this procedure:
   * First create the adabas command, which is the workhorse of this extension.
   */
  Tcl_CreateCommand (interp, "adabas",
		     AdabasCmd, (ClientData) NULL,
		     (Tcl_CmdDeleteProc *) NULL);

  /*
   * Define the place, where the AdabasTcl library goes; there are the
   * sourcable .tcl files - not the shared library, which resides in
   * $DBROOT/lib, since this place you must know to execute this code).
   */
  if (!adabasTclDir || !*adabasTclDir) adabasTclDir = ADABASTCL_DIR;
  Tcl_SetVar (interp, "adabastcl_library", adabasTclDir, TCL_GLOBAL_ONLY);

  /*
   * And here are the API procedures adaopen, adasql, adafetch,...
   * (that are defined in tcl in the first version).
   */
  if (AdabasSqlInit (interp) == TCL_ERROR) {
    fprintf (stderr, "AdabasSqlInit failed: %s\n", interp->result);
  }

#ifdef ADABAS_TCLSH
  Tcl_SetVar(interp, "tcl_rcFileName", "~/.adabastclshrc", TCL_GLOBAL_ONLY);
#endif

#ifdef ADABAS_WISH
  Tcl_SetVar(interp, "tcl_rcFileName", "~/.adabaswishrc",  TCL_GLOBAL_ONLY);
#endif

  sprintf (versionString, "%.1f", ADABAS_TCL_VERSION);
  if (Tcl_PkgProvide (interp, "Adabastcl", versionString) == TCL_ERROR)
    return TCL_ERROR;
#ifndef ADABAS_MODULE
  Tcl_StaticPackage (interp, "Adabastcl", static_Init,
		     (Tcl_PackageInitProc *) NULL);
#endif

  return TCL_OK;
}

/* ----------------- Local procedures --------------------------- */

#ifndef ADABAS_MODULE
static int static_Init (interp)
     Tcl_Interp *interp;		/* Current interpreter. */
{
  return TCL_OK;
}
#endif

/* ------------------------------------------------------------------*/

static int AdabasCmd (clientData, interp, argc, argv)
     ClientData  clientData;		/* Reached through to subcommands. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         argc;			/* Number of arguments. */
     char      **argv;			/* Argument strings. */
{
  size_t         length;
  int            c;

  if (argc < 2) {
    Tcl_AppendResult (interp, "wrong # args: should be \"", argv[0],
		      " option ?arg ...?\"", (char *) NULL);
    return TCL_ERROR;
  }

  c      = argv[1][0];
  length = strlen (argv[1]);

  if (c == 'o' && !strncmp (argv[1], "open", length))
    return AdabasOpenCmd (clientData, interp, argc-1, argv+1); 
  if (c == 'c' && !strncmp (argv[1], "crypt", length))
    return AdabasCryptCmd (clientData, interp, argc-1, argv+1); 
  if (c == 'f' && !strncmp (argv[1], "fapp", length) && length >= 2)
    return AdabasXGetCmd (clientData, interp, 2, argc-1, argv+1); 
  if (c == 'f' && !strncmp (argv[1], "fget", length) && length >= 2)
    return AdabasXGetCmd (clientData, interp, 1, argc-1, argv+1); 
  if (c == 'g' && !strncmp (argv[1], "get", length))
    return AdabasGetCmd (clientData, interp, argc-1, argv+1); 
  if (c == 'p' && !strncmp (argv[1], "put", length))
    return AdabasPutCmd (clientData, interp, argc-1, argv+1); 
  if (c == 's' && !strncmp (argv[1], "sget", length))
    return AdabasXGetCmd (clientData, interp, 0, argc-1, argv+1); 
  if (c == 't' && !strncmp (argv[1], "termId", length)) {
    AdabasRteTermId (interp->result); 
    return TCL_OK;
  }
  if (c == 'v' && !strncmp (argv[1], "version", length))
    return AdabasVersion (clientData, interp, argc-1, argv+1); 
  if (c == 'x' && !strncmp (argv[1], "xuser",   length))
    return AdabasXUser (clientData, interp, argc-1, argv+1); 

  Tcl_AppendResult (interp, "bad option \"", argv[1], "\": should be ",
		    "open, crypt, fget, fapp, get, put, sget, termId, "
		    "version or xuser", (char *) NULL);
  return TCL_ERROR;
}

/* ------------------------------------------------------------------*/

static int AdabasOpenCmd (dummy, interp, argc, argv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         argc;			/* Number of arguments. */
     char      **argv;			/* Argument strings. */
{
  int         service = SQL_USER;
  char       *serverdb;
  char        cmdName[DBNAME+7];
  size_t      length;
  int         c;
  Tcl_CmdInfo cmdInfo;
  AdabasInfo *vars;

  /*
   * Processing of the options.
   */
  serverdb = Tcl_GetVar2 (interp, "env", "DBNAME", TCL_GLOBAL_ONLY);
  *cmdName = '\0';

  for (argv+=1, argc-=1; argc > 1; argv += 2, argc -= 2) {
    c      = argv[0][1];
    length = strlen (argv[0]);

    if (c == 'd' && !strncmp (argv[0], "-dbname", length) && length > 1) {
      serverdb = argv[1];
    } else if (c == 'n' && !strncmp (argv[0], "-name", length) && length > 1) {
      strncpy (cmdName, argv[1], DBNAME);
    } else if (c == 's' && !strncmp (argv[0], "-service", length)
	       && length > 1) {
      if ((service = scanService (interp, argv[1])) < 0)
	return TCL_ERROR;
    } else {
      Tcl_AppendResult (interp, "bad switch \"", argv[0],
			"\": must be -dbname, -name or -service",
			(char *) NULL);
      return TCL_ERROR;
    }
  }
  if (argc > 0) {
    Tcl_AppendResult (interp, "missing argument for switch: \"", argv[0],
		      "\"", (char *) NULL);
    return TCL_ERROR;
  }
  if (!serverdb || !*serverdb) {
    interp->result = "Neither -dbname specified nor env(DBNAME) set";
    return TCL_ERROR;
  }

  /*
   * Initialize the database runtime environment and allocate the client_data.
   */
  if ((vars = (AdabasInfo *) ckalloc (sizeof (AdabasInfo))) == NULL) {
    interp->result = "Initialization failed due to lack of memory";
    return TCL_ERROR;
  }
  AdabasRteInit    (&vars->rteInfo);
  AdabasPacketInit (&vars->packetInfo);
#if ADABAS_VERSION <= 61
  memset (vars->sqlPacket.rte_header, 0, sizeof (rte_header));
#endif
  vars->partInfos[ pkNil         ].subst = 0;
  vars->partInfos[ pkCommand     ].subst = 0;
  vars->partInfos[ pkData        ].subst = 1;
  vars->partInfos[ pkLongdata    ].subst = 0;
  vars->partInfos[ pkResultcount ].subst = 0;
  vars->partInfos[ pkParsid      ].subst = 1;

  /*
   * Now open the connection to the given database.
   */
  if (AdabasRteConnect (interp, &vars->rteInfo,
			serverdb, service) == TCL_ERROR) {
    ckfree ((char *) vars);
    return TCL_ERROR;
  }
  
  /*
   * Evaluate the name of the command to create:
   * If a name is given, use it, else take the name of the serverdb.
   * Append ' #n' with n a small int, if another command with that name
   * already exists.
   */
  if (!*cmdName) strncpy (cmdName, serverdb, DBNAME);
  c = 1;
  while (Tcl_GetCommandInfo (interp, cmdName, &cmdInfo)) {
    sprintf (cmdName, "%.*s #%d", DBNAME, serverdb, ++c);
  }
    
  /*
   * Now create the command with the evaluated name.
   */
  Tcl_CreateCommand (interp, cmdName,
		     AdabasSendCmd, (ClientData) vars,
		     (Tcl_CmdDeleteProc *) NULL);
  Tcl_SetResult (interp, cmdName, TCL_VOLATILE);

  return TCL_OK;
}

/* ------------------------------------------------------------------*/

int AdabasSendCmd (clientData, interp, argc, argv)
     ClientData  clientData;		/* Logon handle (generic type). */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         argc;			/* Number of arguments. */
     char      **argv;			/* Argument strings. */
{
  AdabasInfo      *vars        = (AdabasInfo *) clientData;
  AdabasPacketInfo thisOptions = vars->packetInfo;
  int              partKind;
  size_t           length;
  int              c;
  int              partI;
  char             serverdb[DBNAME+1];
#if ADABAS_VERSION <= 61
  int              orgMode;
  int2             currPos;
#endif

  if (argc < 2) {
    Tcl_AppendResult (interp, "wrong # args: should be \"", argv[0],
		      " option ?arg ...?\"", (char *) NULL);
    return TCL_ERROR;
  }

  c      = argv[1][0];
  length = strlen (argv[1]);

  if (c == 's' && !strncmp (argv[1], "send", length)) {
    /*
     * <command> send: communicate with the database kernel by
     *                 - constructing the order packet,
     *                 - sending it to the kernel and
     *                 - waiting for the answer.
     */

    for (partI = 0; partI < PART_KINDS; partI++) {
      vars->partInfos[ partI ].bufLength = 0;
    }

    /*
     * Processing of the options.
     */
    if (argc < 3) {
      interp->result = "missing specification, what to send";
      return TCL_ERROR;      
    }
    for (argv+=2, argc-=2; argc > 1; argv += 2, argc -= 2) {
      c        = argv[0][1];
      length   = strlen (argv[0]);
      partKind = pkNil;

      if (c == 'c' && !strncmp (argv[0], "-command", length)) {
	partKind = pkCommand;
      } else if (c == 'd' && !strncmp (argv[0], "-data", length)) {
	partKind = pkData;
      } else if (c == 'l' && !strncmp (argv[0], "-longdata", length)) {
	partKind = pkLongdata;
      } else if (c == 'p' && !strncmp (argv[0], "-parsid", length)) {
	partKind = pkParsid;
      } else if (c == 'r' && !strncmp (argv[0], "-resultcount", length)) {
	partKind = pkResultcount;
      } else {
	if (scanPacketConfOpts (interp,
				"-command, -data, -resultcount, -longdata, ", 
				argv[0], argc-1, argv+1,
				&thisOptions) == TCL_ERROR) {
	  return TCL_ERROR;
	}
      }

      if (partKind != pkNil) {
	if (vars->partInfos[ partKind ].bufLength) {
	  Tcl_AppendResult (interp, "\"", argv[0], "\" is given twice",
			    (char *) NULL);
	  return TCL_ERROR;
	}
	vars->partInfos[partKind].buf = argv[1];
	if (partKind == pkResultcount) {
	  if (!sscanf (argv[1], "%d", &vars->partInfos[partKind].bufLength)) {
	    Tcl_AppendResult (interp, "Invalid result count \"",
			      argv[1], "\"", (char *) NULL);
	    return TCL_ERROR;
	  }
	} else {
	  vars->partInfos[partKind].bufLength = strlen (argv[1]);
	}
      }
    }

    if (argc > 0) {
      Tcl_AppendResult (interp, "missing argument for switch: \"", argv[0],
			"\"", (char *) NULL);
      return TCL_ERROR;
    }

    if (AdabasFillPacket (interp, vars, &thisOptions) == TCL_ERROR)
      return TCL_ERROR;

    if (AdabasSendPacket (interp, vars) == TCL_ERROR)
      return TCL_ERROR;
    
    /*
     * Transform the received packet to a valid tcl result.
     */
#if ADABAS_VERSION >= 62
    setResultSegment (interp, &vars->rcvPacket->segm);
#else /* ADABAS_VERSION <= 61 */
    setResultSegment (interp, &vars->sqlPacket, thisOptions.messType);
#endif

#if ADABAS_VERSION <= 61
    /*
     * Since in pre6.2 there is no possibility to set autocommit
     * or withInfo, when switching the sqlmode, we must emulate this.
     * We set sqlmode temporarily to adabas to avoid endless loops...
     */
    if (thisOptions.sqlMode != SQLM_ADABAS) {
      orgMode = vars->packetInfo.sqlMode;
      vars->packetInfo.sqlMode = SQLM_ADABAS;
      if (thisOptions.commitImmediately) {
	if (AdabasSendCmdPacket (interp, vars, 0, 0,
				 "COMMIT WORK") == TCL_ERROR)
	  return TCL_ERROR;
      }
      if (thisOptions.withInfo) {
	if (AdabasSendCmdPacket (interp, vars, 0, 0,
				 "DESCRIBE") == TCL_ERROR)
	  return TCL_ERROR;
	currPos = 1; /* This should be one result part only... */
	while (currPos <= vars->sqlPacket.part1_length)
	  putResultPart (interp, 1, &currPos, &vars->sqlPacket,
			 thisOptions.messType);
      }
      vars->packetInfo.sqlMode = orgMode;
    }
#endif

  } else if (c == 'c' && !strncmp (argv[1], "close", length) && length > 1) {
    /*
     * <command> close: Close a connection to the database and delete
     *                  this tcl command.
     */
    return AdabasCloseCmd (interp, argv[0], vars);

  } else if (c == 'c' && !strncmp (argv[1], "cancel", length) && length > 1) {
    /*
     * <command> cancel: Cancels the currently running sql statement.
     */
    AdabasRteCancel (&vars->rteInfo);

  } else if (c == 'r' && !strncmp (argv[1], "reopen", length)) {
    /*
     * <command> reopen: Close and then open the connection to the database.
     */
    AdabasRteClose (&vars->rteInfo);
    p2c (serverdb, vars->rteInfo.dbname, DBNAME);
    if (AdabasRteConnect (interp, &vars->rteInfo,
			  serverdb, vars->rteInfo.service) == TCL_ERROR)
      return TCL_ERROR;

  } else if (c == 'c' && !strncmp (argv[1], "configure", length)
	     && length > 1) {
    /*
     * <command> configure: Query or set the configuration.
     */

    if (argc == 2) {
      return AdabasSendConfigure (interp, &vars->packetInfo, (char *) NULL);
    } else if (argc == 3) {
      return AdabasSendConfigure (interp, &vars->packetInfo, argv[2]);
    } else {
      return scanPacketConfOpts (interp, "", argv[2], argc-3, argv+3,
				 &vars->packetInfo);
    }
  } else {
    Tcl_AppendResult (interp, "bad option \"", argv[1],
		      "\": should be send, configure, cancel, close or reopen",
		      (char *) NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

/* ------------------------------------------------------------------*/

static int AdabasXGetCmd (dummy, interp, toFile, argc, argv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         toFile;		/* 0: return long value in interp;
					 * 1: write long value in file;
					 * 2: append long value to file. */
     int         argc;			/* Number of arguments. */
     char      **argv;			/* Argument strings. */
{
  int ret;
  int length;
  int paramCnt = toFile ? 4 : 3;
  Tcl_DString     substitute;
  long_descriptor longDesc;
#if ADABAS_VERSION <= 61
  int valpos_corr;
#endif
  
  if (argc != paramCnt) {
    Tcl_AppendResult (interp, "wrong # args: should be \"adabas ",
		      toFile == 1 ? "fget" : toFile == 2 ? "fapp" : "sget",
		      " string descriptor",
		      toFile ? " fileName" : "", "\"", (char *) NULL);
    return TCL_ERROR;
  }

  if (list2LongDescr (interp, argv[2], &longDesc) == TCL_ERROR)
    return TCL_ERROR;

  Tcl_DStringInit (&substitute);

#if ADABAS_VERSION >= 62
  length = substBackslash (&substitute, argv[1],
			   longDesc.valpos,
			   longDesc.vallen);
#else /* ADABAS_VERSION <= 61 */
  /*
   * In pre-6.2 the long descr have positions, that count from the start
   * of part2, and our -data element starts behind the resultcount at
   * position 12, and -longdata elements start behind the two byte
   * integer at position 3. So we must correct the valpos entries.
   *
   * The problem is, that here we haven't the information, what kind of
   * part the given string is from. But we assume, that it is a -data
   * part (returned by a fetch), if the intern pos is exactly on byte
   * behind the vallen, a -longdata part (returned by getval) otherwise.
   */
  if (longDesc.intern_pos == longDesc.vallen + 1)
    valpos_corr = 11; /* returned by a fetch;  offset is resultcount.    */
  else
    valpos_corr = 2;  /* returned by a getval; offset is two byte count. */

  length = substBackslash (&substitute, argv[1],
			   longDesc.valpos - valpos_corr,
			   longDesc.vallen);
#endif

  if (toFile) {
    ret = AdabasLongToFile (interp, toFile, argv[3], length,
			    Tcl_DStringValue (&substitute));
    Tcl_DStringFree (&substitute);
    return ret;
  } else {
    Tcl_DStringResult (interp, &substitute);
    return TCL_OK;
  }
}

/* ------------------------------------------------------------------*/

static int AdabasVersion (dummy, interp, argc, argv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         argc;			/* Number of arguments. */
     char      **argv;			/* Argument strings. */
{
  if (argc != 1) {
    interp->result = "wrong # args: should be adabas version";
    return TCL_ERROR;
  }
  SET_VERSION (interp->result);
  return TCL_OK;
}

/* ------------------------------------------------------------------*/

static int AdabasXUser (dummy, interp, argc, argv)
     ClientData  dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int         argc;			/* Number of arguments. */
     char      **argv;			/* Argument strings. */
{
  size_t       length;
  int          c;
  xuser_record userParams;
  int          index;
  char         givenPasswd[100]; /* more then enough. */

  if (argc == 1) {
    interp->result = "wrong # args: should be adabas xuser option ?...?";
    return TCL_ERROR;
  }
  c      = argv[1][0];
  length = strlen (argv[1]);

  if (c == 'a' && !strncmp (argv[1], "args", length)) {
    if (AdabasXUserArgs (interp, &userParams, givenPasswd) == TCL_ERROR)
      return TCL_ERROR;
    xuser2result (interp, &userParams, givenPasswd);
    return TCL_OK;
  }

  if (c == 'c' && !strncmp (argv[1], "clear", length) && length > 1) {
    return AdabasXUserClear (interp);
  }

  if (c == 'c' && !strncmp (argv[1], "close", length) && length > 1) {
    return AdabasXUserClose (interp);
  }

  if (c == 'g' && !strncmp (argv[1], "get", length)) {
    if (argc > 3) {
      interp->result = "wrong # args: should be adabas xuser get ?userkey?";
      return TCL_ERROR;
    }
    if (AdabasXUserGet (interp, argv[2], &userParams) == TCL_ERROR)
      return TCL_ERROR;
    xuser2result (interp, &userParams, (char *) NULL);
    return TCL_OK;
  }

  if (c == 'i' && !strncmp (argv[1], "index", length)) {
    if (argc != 3) {
      interp->result = "wrong # args: should be adabas xuser index no";
      return TCL_ERROR;
    }
    if (Tcl_GetInt (interp, argv[2], &index) == TCL_ERROR)
      return TCL_ERROR;
    if (AdabasXUserIndex (interp, (int2) index, &userParams) == TCL_ERROR)
      return TCL_ERROR;
    xuser2result (interp, &userParams, (char *) NULL);
    return TCL_OK;
  }

  if (c == 'o' && !strncmp (argv[1], "open", length)) {
    return AdabasXUserOpen (interp);
  }

  if (c == 'p' && !strncmp (argv[1], "put", length)) {
    if (argc == 2) {
      interp->result =
	"wrong # args: adabas xuser put entry value ?entry value?...";
      return TCL_ERROR;
    }
    if (args2xuser (interp, argc-2, argv+2, &userParams) == TCL_ERROR)
      return TCL_ERROR;
    return AdabasXUserPut (interp, &userParams);
  }

  Tcl_AppendResult (interp, "bad option \"", argv[1],
		    "\": must be args, clear, close, get, index, open or put",
		    (char *) NULL);
  return TCL_ERROR;
}

/* ------------------------------------------------------------------*/

static void xuser2result (interp, userParams, givenPasswd)
     Tcl_Interp   *interp;		/* Current interpreter. */
     xuser_record *userParams;		/* Incoming xuser record. */
     char         *givenPasswd;		/* Clear passwd from command line. */
{
  char        buffer[256];
  Tcl_DString pw;

  Tcl_AppendElement (interp, "userkey");
  p2c (buffer, userParams->xu_key,
       sizeof (userParams->xu_key));
  Tcl_AppendElement (interp, buffer);

  Tcl_AppendElement (interp, "servernode");
  p2c (buffer, userParams->xu_servernode,
       sizeof (userParams->xu_servernode));
  Tcl_AppendElement (interp, buffer);

  Tcl_AppendElement (interp, "serverdb");
  p2c (buffer, userParams->xu_serverdb,
       sizeof (userParams->xu_serverdb));
  Tcl_AppendElement (interp, buffer);

  Tcl_AppendElement (interp, "user");
  p2c (buffer, userParams->xu_user,
       sizeof (userParams->xu_user));
  Tcl_AppendElement (interp, buffer);

  if (givenPasswd && *givenPasswd) {
    Tcl_AppendElement (interp, "clearpw");
    Tcl_AppendElement (interp, givenPasswd);
  } else {
    Tcl_AppendElement (interp, "password");
    memcpy (buffer, userParams->xu_password,
	    sizeof (userParams->xu_password));
    Tcl_DStringInit (&pw);
    escapeSequence (&pw, buffer, 24);
    Tcl_AppendElement (interp, Tcl_DStringValue (&pw));
    Tcl_DStringFree (&pw);
  }

  Tcl_AppendElement (interp, "sqlmode");
  p2c (buffer, userParams->xu_sqlmode,
       sizeof (userParams->xu_sqlmode));
  Tcl_AppendElement (interp, buffer);

#if ADABAS_VERSION <= 61
  Tcl_AppendElement (interp, "cachelimit");
  sprintf (buffer, "%d", userParams->xu_cachelimit);
  Tcl_AppendElement (interp, buffer);
#endif

  Tcl_AppendElement (interp, "timeout");
  sprintf (buffer, "%d", userParams->xu_timeout);
  Tcl_AppendElement (interp, buffer);

  Tcl_AppendElement (interp, "isolation");
  sprintf (buffer, "%d", userParams->xu_isolation);
  Tcl_AppendElement (interp, buffer);

#if ADABAS_VERSION >= 62
  Tcl_AppendElement (interp, "dblang");
  p2c (buffer, userParams->xu_dblang,
       sizeof (userParams->xu_dblang));
  Tcl_AppendElement (interp, buffer);
#endif
}

/* ------------------------------------------------------------------*/

static int args2xuser (interp, argc, argv, userParams)
     Tcl_Interp   *interp;		/* Current interpreter. */
     int           argc;		/* Number of arguments. */
     char        **argv;		/* Argument strings. */
     xuser_record *userParams;		/* Returning xuser record. */
{
  size_t      length;
  int         c;
  Tcl_DString pw;
  int         value;
  int         keyFound = 0;

  /*
   * First initialize all char fields to blank and numeric fileds to -1.
   */
  memset (userParams, ' ', sizeof (*userParams));
  c2p (userParams->xu_sqlmode, sqlMode2String (SQLM_ADABAS, 1),
       sizeof (userParams->xu_sqlmode));
  userParams->xu_cachelimit = -1;
  userParams->xu_timeout    = -1;
  userParams->xu_isolation  = -1;

  /*
   * Now scan in a loop all given parameter specs.
   */
  while (argc >= 2) {
    c      = argv[0][0];
    length = strlen (argv[0]);

    switch (c) {
    case 'c':
      if (!strncmp (argv[0], "cachelimit", length)) {
	if (Tcl_GetInt (interp, argv[1], &value) == TCL_ERROR)
	  return TCL_ERROR;
	userParams->xu_cachelimit = (int4) value;
	break;
      }
#if ADABAS_VERSION >= 62
    case 'd':
      if (!strncmp (argv[0], "dblang", length)) {
	c2p (userParams->xu_dblang, argv[1],
	     sizeof (userParams->xu_dblang));
	break;
      }
#endif
    case 'i':
      if (!strncmp (argv[0], "isolation", length)) {
	if (Tcl_GetInt (interp, argv[1], &value) == TCL_ERROR)
	  return TCL_ERROR;
	userParams->xu_isolation = (int2) value;
	break;
      }
    case 's':
      if (!strncmp (argv[0], "servernode", length) && length > 6) {
	c2p (userParams->xu_servernode, argv[1],
	     sizeof (userParams->xu_servernode));
	break;
      }
      if (!strncmp (argv[0], "serverdb", length) && length > 6) {
	c2p (userParams->xu_serverdb, argv[1],
	     sizeof (userParams->xu_serverdb));
	break;
      }
      if (!strncmp (argv[0], "sqlmode", length) && length > 1) {
	if (scanSqlMode (interp, argv[1], 1) < 0)
	  return TCL_ERROR;
	c2p (userParams->xu_sqlmode, argv[1],
	     sizeof (userParams->xu_sqlmode));
	break;
      }
    case 'p':
      if (!strncmp (argv[0], "password", length)) {
	Tcl_DStringInit (&pw);
	(void) substBackslash (&pw, argv[1], 1, 0);
	memcpy (userParams->xu_password, Tcl_DStringValue (&pw),
		sizeof (userParams->xu_password));
	Tcl_DStringFree (&pw);
	break;
      }
    case 't':
      if (!strncmp (argv[0], "timeout", length)) {
	if (Tcl_GetInt (interp, argv[1], &value) == TCL_ERROR)
	  return TCL_ERROR;
	userParams->xu_cachelimit = (int2) value;
	break;
      }
    case 'u':
      if (!strncmp (argv[0], "user", length) && length == 4) {
	c2p (userParams->xu_user, argv[1],
	     sizeof (userParams->xu_user));
	break;
      }
      if (!strncmp (argv[0], "userkey", length) && length > 4) {
	c2p (userParams->xu_key, argv[1],
	     sizeof (userParams->xu_key));
	keyFound = 1;
	break;
      }
    default:
      Tcl_AppendResult (interp, "Invalid xuser entry \"", argv[0],
			"\"", (char *) NULL);
      return TCL_ERROR;
    }
    argc -= 2;
    argv += 2;
  }

  /*
   * Is there a dangling parameter on the command line?
   */
  if (argc == 1) {
    Tcl_AppendResult (interp, "Expected value for entry \"", argv[0],
		      "\"", (char *) NULL);
    return TCL_ERROR;
  }

  /*
   * One of the specifications must be the userkey itself.
   */
  if (!keyFound) {
    interp->result =  "No userkey specified";
    return TCL_ERROR;
  }
  return TCL_OK;
}

/* ------------------------------------------------------------------*/

int AdabasCloseCmd (interp, cmdName, vars)
     Tcl_Interp *interp;		/* Current interpreter. */
     char       *cmdName;		/* Name of logon handle to delete. */
     AdabasInfo *vars;			/* Logon handle to delete. */
{
  /*
   * Delete this tcl command.
   */
  if (Tcl_DeleteCommand (interp, cmdName) == TCL_ERROR) {
    Tcl_AppendResult (interp, "can't delete command \"", cmdName,
		      "\"", (char *) NULL);
    return TCL_ERROR;
  }

  /*
   * Close the connection to the database.
   */
  AdabasRteClose (&vars->rteInfo);
  ckfree ((char *) vars);
  return TCL_OK;
}

/* ------------------------------------------------------------------*/

static int AdabasSendConfigure (interp, vars, optionName)
     Tcl_Interp       *interp;		/* Current interpreter. */
     AdabasPacketInfo *vars;		/* Packet handle. */
     char             *optionName;	/* Name of option to examine. */
{
  int         length = optionName ? strlen (optionName) : 0;
  OptionMap  *o;
  Tcl_DString oneValue;
  char       *defaultValue;
  char       *currentValue;

  Tcl_DStringInit (&oneValue);
  for (o = sendOptions; o->name; o++) {
    if (!optionName || strncmp (optionName, o->name, length) == 0) {
      switch (o->kind) {

      case oWithInfo:
	defaultValue = "false";
	currentValue = vars->withInfo ? "true" : "false";
	break;

      case oMessType:
	defaultValue = messType2String (M_DBS);
	currentValue = messType2String (vars->messType);
	break;

      case oProducer:
#if ADABAS_VERSION >= 62
        defaultValue = producer2String (PR_USER_CMD);
	currentValue = producer2String (vars->producer);
#else /* ADABAS_VERSION <= 61 */
        defaultValue = currentValue = "user";
#endif
	break;

      case oSqlMode:
#if ADABAS_VERSION >= 62
	defaultValue = sqlMode2String (SM_SESSION_SQLMODE, 0);
#else /* ADABAS_VERSION <= 61 */
	defaultValue = sqlMode2String (SQLM_ADABAS, 0);
#endif
	currentValue = sqlMode2String (vars->sqlMode, 0);
	break;

      case oMassCmd:
	defaultValue = "false";
	currentValue = vars->massCmd ? "true" : "false";
	break;

      case oCommitImmediately:
	defaultValue = "false";
	currentValue = vars->commitImmediately ? "true" : "false";
	break;

      case oApplication:
	defaultValue = application2String ("XCI");
	currentValue = application2String (vars->application);
	break;

      default:
	interp->result = "Oops, invalid configure option?!?";
	return TCL_ERROR;
      }

      Tcl_DStringAppendElement (&oneValue, o->name);
      Tcl_DStringAppendElement (&oneValue, defaultValue);
      Tcl_DStringAppendElement (&oneValue, currentValue);

      if (optionName) {
	Tcl_DStringResult (interp, &oneValue);
	return TCL_OK;
      } else {
	Tcl_AppendElement (interp, Tcl_DStringValue (&oneValue));
	Tcl_DStringFree (&oneValue);
      }
    }
  }

  if (optionName) {
    Tcl_AppendResult (interp, "bad switch \"", optionName,
		      "\": must be -messType, -producer, -sqlMode",
		      " -massCmd, -commitImmediately or -withInfo",
		      (char *) NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

/* ------------------------------------------------------------------*/

static int scanPacketConfOpts (interp, expected, optionName, argc, argv, vars)
     Tcl_Interp       *interp;		/* Current interpreter. */
     char             *expected;	/* List of options (for errortext). */
     char             *optionName;	/* Name of option to configure. */
     int               argc;		/* Number of arguments. */
     char            **argv;		/* Argument strings. */
     AdabasPacketInfo *vars;		/* Packet handle. */
{
  int   c      = optionName[1];
  int   length = strlen (optionName);
  int   newMessType;
  char *newApplication;
  int   newSqlMode;
#if ADABAS_VERSION >= 62
  int   newProducer;
#endif

  /*
   * <cmd> configure -application <commandInterface|load|utility...>
   * Kind of application, sending the request to the kernel.
   */
  if (c == 'a' && !strncmp (optionName, "-application", length)) {
    if (!(newApplication = scanApplication (interp, argv[0])))
      return TCL_ERROR;
    memcpy (vars->application, newApplication, 3);
    return TCL_OK;
  }

  /*
   * <cmd> configure -messType <dbs|parse|execute|utility|...>
   * Message type of packet, that will be send to the kernel.
   */
  if (c == 'm' && !strncmp (optionName, "-messType", length) && length > 2) {
    newMessType = scanMessType (interp, argv[0]);
    if (newMessType == M_NIL) return TCL_ERROR;
    vars->messType = newMessType;
    return TCL_OK;
  }

#if ADABAS_VERSION >= 62
  /*
   * <cmd> configure -producer <user|internal|installation>
   * Who is the producer of the request?
   */
  if (c == 'p' && !strncmp (optionName, "-producer", length)) {
    if ((newProducer = scanProducer (interp, argv[0])) == PR_NIL)
      return TCL_ERROR;
    vars->producer = newProducer;
    return TCL_OK;
  }
#endif

  /*
   * <cmd> configure -sqlMode <session|adabas|ansi|db2|oracle>
   * In what mode should the request be interpreted?
   */
  if (c == 's' && !strncmp (optionName, "-sqlMode", length)) {
#if ADABAS_VERSION >= 62
    if ((newSqlMode = scanSqlMode (interp, argv[0], 0)) == SM_NIL)
#else /* ADABAS_VERSION <= 61 */
    if ((newSqlMode = scanSqlMode (interp, argv[0], 0)) < 0)
#endif
      return TCL_ERROR;
    vars->sqlMode = newSqlMode;
    return TCL_OK;
  }

  /*
   * <cmd> configure -withInfo <bool>.
   * Should the kernel respond with the addition info of columnnames
   * and shortinfos for a select?
   */
  if (c == 'w' && !strncmp (optionName, "-withInfo", length)) {
    return Tcl_GetBoolean (interp, argv[0], &vars->withInfo);
  }

  /*
   * <cmd> configure -massCmd <bool>.
   * Should the given sqlStatement be executed as mass command?
   */
  if (c == 'm' && !strncmp (optionName, "-massCmd", length) && length > 2) {
    return Tcl_GetBoolean (interp, argv[0], &vars->massCmd);
  }

  /*
   * <cmd> configure -commitImmediately <bool>.
   * Should the kernel execute a commit after any ddl statement?
   */
  if (c == 'c' && !strncmp (optionName, "-commitImmediately", length)) {
    return Tcl_GetBoolean (interp, argv[0], &vars->commitImmediately);
  }

  /*
   * No known option was given, if control gets to this point.
   */
  Tcl_AppendResult (interp, "bad switch \"", optionName, "\": must be ",
		    expected, "-application, -messType, -producer,",
		    " -sqlMode, -masscmd, -commitImmediately or -withInfo",
		    (char *) NULL);
  return TCL_ERROR;
}

/* ------------------------------------------------------------------*/

#if ADABAS_VERSION >=62
static void fillPacketHeader (head, size, swapKind, application)
     packet_header *head;		/* Header of packet to be filled. */
     int            size;		/* Size of packet. */
     int            swapKind;		/* Swap kind (see adabas.h). */
     char          *application;	/* Application. */
{
  head->mess_code    = 0; /*is_ascii*/;
  head->mess_swap    = swapKind;
  head->varpart_size = size - sizeof (packet_header);
  head->varpart_len  = 0;
  head->no_of_segm   = 0;
  memcpy (head->appl_version, "62000",     5);
  memcpy (head->application,  application, 3);
}
#else /* ADABAS_VERSION <= 61 */
static void fillPacketHeader (head, swapKind, application)
     packet *head;			/* Packet with header to be filled. */
     int     swapKind;			/* Swap kind (see adabas.h). */
     char   *application;		/* Application. */
{
  head->mess_code_type = 0; /*is_ascii*/;
  head->mess_swap_kind = swapKind;
  head->return_code    = 0;
  head->error_code     = 0;
  head->part1_length   = 0;
  head->part2_length   = 0;
  memcpy (head->senderid,   "61115",     5);
  memcpy (head->senderid+5, application, 3);
}
#endif

/* ------------------------------------------------------------------*/

#if ADABAS_VERSION >=62
static void fillSegmentHeader (packetHead, segmentHead, options)
     packet_header    *packetHead;	/* Header of packet. */
     segment_header   *segmentHead;	/* Header of segment to be filled. */
     AdabasPacketInfo *options;		/* Current packet options. */
{
  segmentHead->kind.cmd.segm_len            = sizeof (segment_header);
  segmentHead->kind.cmd.segm_offset         = 0;
  segmentHead->kind.cmd.no_of_parts         = 0;
  segmentHead->kind.cmd.own_index           = 1;
  segmentHead->kind.cmd.segm_kind           = SK_CMD;
  segmentHead->kind.cmd.mess_type           = options->messType;
  segmentHead->kind.cmd.sqlmode             = options->sqlMode;
  segmentHead->kind.cmd.producer            = options->producer;
  segmentHead->kind.cmd.commit_immediately  = options->commitImmediately;
  segmentHead->kind.cmd.ignore_costwarning  = 0;
  segmentHead->kind.cmd.prepare             = 0;
  segmentHead->kind.cmd.with_info           = options->withInfo;
  segmentHead->kind.cmd.mass_cmd            = options->massCmd;

  packetHead->varpart_len += sizeof (segment_header);
  packetHead->no_of_segm  += 1;
}
#else /* ADABAS_VERSION <= 61 */
static void fillSegmentHeader (packetHead, options)
     packet           *packetHead;	/* Packet with segment to be filled. */
     AdabasPacketInfo *options;		/* Current packet options. */
{
  packetHead->mess_type = options->messType;
  if (options->commitImmediately)
    switch (packetHead->mess_type) {
    case M_DBS:
      packetHead->mess_type = M_DBSCOMMIT;
      break;
    case M_EXECUTE:
      packetHead->mess_type = M_DBSCOMMIT_EXECUTE;
      break;
    }
  if (options->withInfo)
    switch (packetHead->mess_type) {
    case M_DBS:
      packetHead->mess_type = M_DBSINFO;
      break;
    case M_DBSCOMMIT:
      packetHead->mess_type = M_DBSINFOCOMMIT;
      break;
    }
  switch (options->sqlMode) {
  case SQLM_ANSI:
    switch (packetHead->mess_type) {
    case M_DBS:
    case M_DBSINFO:
    case M_DBSCOMMIT:
    case M_DBSINFOCOMMIT:
      packetHead->mess_type = M_DBSANSI;
    case M_PARSE:
      packetHead->mess_type = M_PARSEANSI;
    }
    break;
  case SQLM_DB2:
    switch (packetHead->mess_type) {
    case M_DBS:
    case M_DBSINFO:
    case M_DBSCOMMIT:
    case M_DBSINFOCOMMIT:
      packetHead->mess_type = M_DBSDB2;
    case M_PARSE:
      packetHead->mess_type = M_PARSEDB2;
    }
    break;
  case SQLM_ORACLE:
    switch (packetHead->mess_type) {
    case M_DBS:
    case M_DBSINFO:
    case M_DBSCOMMIT:
    case M_DBSINFOCOMMIT:
      packetHead->mess_type = M_DBSORACLE;
    case M_PARSE:
      packetHead->mess_type = M_PARSEORACLE;
    }
    break;
  }
}
#endif

/* ------------------------------------------------------------------*/

#if ADABAS_VERSION >=62
static int fillPart (interp, packetHead, segmentHead, currPart, kind, partInfo)
     Tcl_Interp     *interp;		/* Current interpreter. */
     packet_header  *packetHead;	/* Header of packet. */
     segment_header *segmentHead;	/* Header of segment. */
     part          **currPart;		/* Part to be filled. */
     int             kind;		/* Part kind (see adabas.h). */
     AdabasPartInfo *partInfo;		/* Current part options. */
{
  part_header *partHead;
  int4         partLength;
  int4         available;
  Tcl_DString  substitute;
  num_error    numError;

  if (partInfo->subst && partInfo->buf) {
    partInfo->bufLength = substBackslash (&substitute, partInfo->buf, 1, 0);
    partInfo->buf       = Tcl_DStringValue (&substitute);
  }
  partHead   = &(*currPart)->part_header;
  available  = packetHead->varpart_size - segmentHead->kind.cmd.segm_len;
  partLength = ((partInfo->bufLength+7)/8)*8 + sizeof (part_header);

  if (partLength > available) {
    if (partInfo->subst) Tcl_DStringFree (&substitute);
    interp->result = "too big packet";
    return TCL_ERROR;
  }

  partHead->part_kind   = kind;
  partHead->attributes  = 0;
  partHead->arg_count   = 1;
  partHead->segm_offset = 0;
  partHead->buf_len     = partInfo->bufLength;
  partHead->buf_size    = available - sizeof (part_header);

  if (kind == PK_RESULTCOUNT) {
    (*currPart)->buf[0] = 0;
    s41plint ((*currPart)->buf, 2, 18, 0, partInfo->bufLength, &numError);
    if (numError) {
      Tcl_AppendResult (interp, "invalid resultcount \"",
			partInfo->buf, "\"", (char *) NULL);
      return TCL_ERROR;
    }
  } else if (partInfo->buf) {
    memcpy ((*currPart)->buf, partInfo->buf, partInfo->bufLength);
  }

  if (partInfo->subst && partInfo->buf)
    Tcl_DStringFree (&substitute);

  packetHead->varpart_len           += partLength;
  segmentHead->kind.cmd.segm_len    += partLength;
  segmentHead->kind.cmd.no_of_parts += 1;

  *currPart = (part *) (((char *) *currPart) + partLength);

  return TCL_OK;
}
#else /* ADABAS_VERSION <= 61 */
static int fillPart (interp, packetHead, kind, partInfo, massCmd)
     Tcl_Interp     *interp;		/* Current interpreter. */
     packet         *packetHead;	/* Packet with part to be filled. */
     partKind        kind;		/* Part kind. */
     AdabasPartInfo *partInfo;		/* Current part options. */
     int             massCmd;		/* Is this a mass command? */
{
  int4        available;
  Tcl_DString substitute;
  int2        argCount;

  if (partInfo->subst && partInfo->buf) {
    partInfo->bufLength = substBackslash (&substitute, partInfo->buf, 1, 0);
    partInfo->buf       = Tcl_DStringValue (&substitute);
  }
  available  = sizeof (packetHead->var_part)
    - packetHead->part1_length - packetHead->part2_length;

  if (partInfo->bufLength > available) {
    if (partInfo->subst) Tcl_DStringFree (&substitute);
    interp->result = "too big packet";
    return TCL_ERROR;
  }

  switch (kind) {
  case pkCommand:
    if (massCmd) {
      packetHead->var_part[0]  = 'M';
      packetHead->part1_length = 1;
    } else
      packetHead->part1_length = 0;
    memcpy (packetHead->var_part + packetHead->part1_length,
	    partInfo->buf, partInfo->bufLength);
    packetHead->part1_length += partInfo->bufLength;
    break;

  case pkData:
    memcpy (packetHead->var_part
	    + packetHead->part1_length + packetHead->part2_length,
	    partInfo->buf, partInfo->bufLength);
    packetHead->part2_length += partInfo->bufLength;
    break;

  case pkResultcount:
    argCount = (int2) partInfo->bufLength;
    memcpy (packetHead->var_part + packetHead->part1_length,
	    (char *) &argCount, 2);
    packetHead->part2_length = 2;
    break;

  case pkLongdata:
    memcpy (packetHead->var_part, partInfo->buf, partInfo->bufLength);
    packetHead->part1_length = 0;
    packetHead->part2_length = partInfo->bufLength;
    break;

  case pkParsid:
    if (partInfo->bufLength != 12) {
      interp->result = "invalid parsid length ???";
      return TCL_ERROR;
    }
    memcpy (packetHead->var_part+packetHead->part1_length, partInfo->buf, 12);
    if (packetHead->part1_length)
      packetHead->part2_length = 12;
    else
      packetHead->part1_length = 12;
    break;
  default:
    if (partInfo->subst) Tcl_DStringFree (&substitute);
    interp->result = "unknown part kind ???";
    return TCL_ERROR;
  }

  if (partInfo->subst && partInfo->buf) Tcl_DStringFree (&substitute);

  return TCL_OK;
}
#endif

/* ------------------------------------------------------------------*/

#if ADABAS_VERSION >= 62
static void setResultSegment (interp, retSegment)
     Tcl_Interp *interp;		/* Current interpreter. */
     segment    *retSegment;		/* Return segment to be converted. */
{
  part *currPart = &retSegment->part;
  int4  partLength;
  char  result[80];
  int   ix;

  sprintf (result, "-returncode %d ",
	   retSegment->segm_header.kind.ret.returncode);
  Tcl_AppendResult (interp, result, (char *) NULL);
  sprintf (result, "-errorpos %d ",
	   retSegment->segm_header.kind.ret.errorpos);
  Tcl_AppendResult (interp, result, (char *) NULL);
  sprintf (result, "-sqlstate %5.5s ",
	   retSegment->segm_header.kind.ret.sqlstate);
  Tcl_AppendResult (interp, result, (char *) NULL);

  for (ix = 0; ix < retSegment->segm_header.kind.cmd.no_of_parts; ix++) {
    if (ix > 0) Tcl_AppendResult (interp, " ", (char *) NULL);
    putResultPart (interp, currPart, &partLength);
    currPart = (part *) (((char *) currPart) + partLength);
  }
}
#else /* ADABAS_VERSION <= 61 */
static void setResultSegment (interp, retSegment, sndMessType)
     Tcl_Interp *interp;		/* Current interpreter. */
     packet     *retSegment;		/* Return segment to be converted. */
     int         sndMessType;		/* Mess type of sended packet. */
{
  char result[80];
  int2 currPos;

  sprintf (result, "-returncode %d ",  retSegment->return_code);
  Tcl_AppendResult (interp, result, (char *) NULL);
  if (retSegment->return_code) {
    sprintf (result, "-errorpos %d ",   retSegment->error_code);
    Tcl_AppendResult (interp, result, (char *) NULL);
  } else {
    /* Warning set not yet... */
  }
  currPos = 1;
  while (currPos <= retSegment->part1_length) {
    putResultPart (interp, 1, &currPos, retSegment, sndMessType);
  }
  currPos = 1;
  while (currPos <= retSegment->part2_length) {
    putResultPart (interp, 2, &currPos, retSegment, sndMessType);
  }
}
#endif

/* ------------------------------------------------------------------*/

int AdabasSendPacket (interp, vars)
     Tcl_Interp *interp;		/* Current interpreter. */
     AdabasInfo *vars;			/* Current logon. */
{
  int      retc;
  char     rc[10];
  char     errt[150];
  int      errp;
#if ADABAS_VERSION >= 62
  segment *retSegm;
#endif

  /*
   * Send the constructed packet to the database and receive a reply.
   */
#if ADABAS_VERSION >= 62
  if (AdabasRteSend (interp, &vars->rteInfo,
		     vars->sndPacket, &vars->rcvPacket) == TCL_ERROR)
    return TCL_ERROR;
#else /* ADABAS_VERSION <= 61 */
  if (AdabasRteSend (interp, &vars->rteInfo, &vars->sqlPacket) == TCL_ERROR)
    return TCL_ERROR;
#endif

#if ADABAS_VERSION >= 62
  retSegm = &vars->rcvPacket->segm;
  if ((retc = retSegm->segm_header.kind.ret.returncode)) {
    errp = retSegm->segm_header.kind.ret.errorpos;
    /*
     * In some strange situations the database kernel returns no part at all.
     * Then we better don't apply onto the error message in part1.
     */
    if (retSegm->segm_header.kind.cmd.no_of_parts)
      p2c (errt, retSegm->part.buf, retSegm->part.part_header.buf_len);
    else
      sprintf (errt, "Error %d", retSegm->segm_header.kind.ret.returncode);
  }
#else /* ADABAS_VERSION <= 61 */
  if ((retc = vars->sqlPacket.return_code)) {
    p2c (errt, vars->sqlPacket.var_part, vars->sqlPacket.part1_length);
    errp = vars->sqlPacket.error_code;
  }
#endif

  if (retc) {
    sprintf (rc, "%d", retc);
    Tcl_SetVar2 (interp, ADABAS_MESSAGES, "rc",       rc,   TCL_GLOBAL_ONLY);
    Tcl_SetResult (interp, rc, TCL_VOLATILE);
    sprintf (rc, "%d", errp);
    Tcl_SetVar2 (interp, ADABAS_MESSAGES, "errorpos", rc,   TCL_GLOBAL_ONLY);
    Tcl_SetVar2 (interp, ADABAS_MESSAGES, "errortxt", errt, TCL_GLOBAL_ONLY);  
    return TCL_ERROR;
  }

  Tcl_SetVar2   (interp, ADABAS_MESSAGES, "rc", "0",  TCL_GLOBAL_ONLY);
  Tcl_UnsetVar2 (interp, ADABAS_MESSAGES, "errortxt", TCL_GLOBAL_ONLY);
  Tcl_UnsetVar2 (interp, ADABAS_MESSAGES, "errorpos", TCL_GLOBAL_ONLY);
  return TCL_OK;
}

/* ------------------------------------------------------------------*/

int AdabasFillPacket (interp, vars, thisOptions)
     Tcl_Interp       *interp;		/* Current interpreter. */
     AdabasInfo       *vars;		/* Current logon. */
     AdabasPacketInfo *thisOptions;	/* Packet options for this command. */
{
  int partKind;
  int partI;

#if ADABAS_VERSION >= 62
  segment  *cmdSegm;
  part     *currPart;
  part_kind kind62;

  /*
   * Initialize the packet header.
   */
  vars->sndPacket = *vars->rteInfo.sqlPacketList;
  fillPacketHeader (&vars->sndPacket->header,
		    vars->rteInfo.sqlPacketSize,
		    vars->packetInfo.swapKind,
		    thisOptions->application);

  /*
   * Initialize the first (and only) command segment.
   */
  cmdSegm = (segment *) &vars->sndPacket->segm;
  fillSegmentHeader (&vars->sndPacket->header,
		     &cmdSegm->segm_header, thisOptions);

  /*
   * Fill the command part with the given string.
   */
  currPart = (part *) &cmdSegm->part;
  for (partKind = partKinds[partI = 0]; partKind;
       partKind = partKinds[++partI])
    if (vars->partInfos[partKind].bufLength) {
      switch (partKind) {
      case pkCommand:     kind62 = PK_COMMAND;       break;
      case pkData:        kind62 = PK_DATA;          break;
      case pkResultcount: kind62 = PK_RESULTCOUNT;   break;
      case pkLongdata:    kind62 = PK_LONGDATA;      break;
      case pkParsid:      kind62 = PK_PARSID;        break;
      default: interp->result = "invalid part kind???"; return TCL_ERROR;
      }
      if (fillPart (interp, &vars->sndPacket->header,
		    &cmdSegm->segm_header, &currPart,
		    kind62, vars->partInfos+partKind) == TCL_ERROR)
	return TCL_ERROR;
    }

#else /* ADABAS_VERSION <= 61 */

  /*
   * Initialize the packet header.
   */
  fillPacketHeader  (&vars->sqlPacket, vars->packetInfo.swapKind,
		     thisOptions->application);
  fillSegmentHeader (&vars->sqlPacket, thisOptions);

  /*
   * Fill all the given parts.
   */
  for (partKind = partKinds[partI = 0]; partKind;
       partKind = partKinds[++partI])
    if (vars->partInfos[partKind].bufLength) {
      if (fillPart (interp, &vars->sqlPacket, partKind,
		    vars->partInfos+partKind,
		    thisOptions->massCmd) == TCL_ERROR)
	return TCL_ERROR;
    }
#endif

  return TCL_OK;
}

/* ------------------------------------------------------------------*/

int AdabasSend2PartPacket (interp, vars, part1Kind, part1, part1Int,
			   part2Kind, part2, part2Int)
     Tcl_Interp *interp;		/* Current interpreter. */
     AdabasInfo *vars;			/* Current logon. */
     int         part1Kind;		/* Part1: Part kind. */
     char       *part1;			/*        Buffer... */
     int         part1Int;		/*        ... or Number. */
     int         part2Kind;		/* Part2: Part kind. */
     char       *part2;			/*        Buffer... */
     int         part2Int;		/*        ... or Number. */
{
  int partI;

  for (partI = 0; partI < PART_KINDS; partI++) {
    vars->partInfos[ partI ].bufLength = 0;
  }
  switch (part1Kind) {
  case pkNil:
    break;
  case pkResultcount:
    vars->partInfos[part1Kind].buf       = "";
    vars->partInfos[part1Kind].bufLength = part1Int;
    break;
  default:
    vars->partInfos[part1Kind].buf       = part1;
    vars->partInfos[part1Kind].bufLength = strlen (part1);
  }
  switch (part2Kind) {
  case pkNil:
    break;
  case pkResultcount:
    vars->partInfos[part2Kind].buf       = "";
    vars->partInfos[part2Kind].bufLength = part2Int;
    break;
  default:
    vars->partInfos[part2Kind].buf       = part2;
    vars->partInfos[part2Kind].bufLength = strlen (part2);
  }

  if (AdabasFillPacket (interp, vars, &vars->packetInfo) == TCL_ERROR)
    return TCL_ERROR;

  return AdabasSendPacket (interp, vars);
}

/* ------------------------------------------------------------------*/

int AdabasSendCmdPacket (interp, vars, withInfo, massCnt, command)
     Tcl_Interp *interp;		/* Current interpreter. */
     AdabasInfo *vars;			/* Current logon. */
     int         withInfo;		/* SQL, that returns results? */
     int         massCnt;		/* Number of rows for mass fetch. */
     char       *command;		/* SQL command to be send. */
{
  int ret;
  int orgWithInfo = vars->packetInfo.withInfo;
  int orgMassCmd  = vars->packetInfo.massCmd;

  vars->packetInfo.withInfo = withInfo;
  vars->packetInfo.massCmd  = massCnt > 1;

  ret = AdabasSend2PartPacket (interp, vars, pkCommand, command, 0,
			       massCnt > 1 ? pkResultcount : pkNil,
			       "", massCnt);

  vars->packetInfo.withInfo = orgWithInfo;
  vars->packetInfo.massCmd  = orgMassCmd;
  return ret;
}

/* ------------------------------------------------------------------*/

int AdabasSendValPacket (interp, vars, wantedMessType, longDesc)
     Tcl_Interp      *interp;		/* Current interpreter. */
     AdabasInfo      *vars;		/* Current logon. */
     int              wantedMessType;	/* Mess type (getval/putval). */
     long_descriptor *longDesc;		/* Long descriptor to send. */
{
  int partI;
  AdabasPacketInfo thisOptions = vars->packetInfo;
  char  descBuf[100];
#if ADABAS_VERSION <= 61
  int2 argCnt      = 1;
#endif

  for (partI = 0; partI < PART_KINDS; partI++) {
    vars->partInfos[ partI ].bufLength = 0;
  }
  
#if ADABAS_VERSION >= 62
  descBuf[0] = 0; /* defined byte */
  memcpy (descBuf+1, (char *) longDesc, 40);
  vars->partInfos[ pkLongdata ].buf       = descBuf;
  vars->partInfos[ pkLongdata ].bufLength = 41;
#else /* ADABAS_VERSION <= 61 */
  memcpy (descBuf,   &argCnt,  2);
  descBuf[2] = 0; /* defined byte */
  memcpy (descBuf+3, longDesc, 32);
  vars->partInfos[ pkLongdata ].buf       = descBuf;
  vars->partInfos[ pkLongdata ].bufLength = 35;
#endif

  thisOptions.messType = wantedMessType;
  if (AdabasFillPacket (interp, vars, &thisOptions) == TCL_ERROR)
    return TCL_ERROR;

  if (AdabasSendPacket (interp, vars) == TCL_ERROR)
    return TCL_ERROR;

  return AdabasLongDescriptor (interp, vars, longDesc, 1, longSendVal);
}
