/* -------------------------------------------------------------- */
/*  LAST EDIT: Sat Aug 17 18:47:07 1996 by Adabas D Database (gutamil.v12-berlin.de!adabas)  */
/*  $Id: AdabasUtil.c,v 1.11 1996/08/17 16:47:13 adabas Exp $ */
/* -------------------------------------------------------------- */
/* Copyright (c) 1996 Christian Krone. All rights reserved.       */
/* This program is free software; you can redistribute it and/or  */
/* modify it under the same terms as Tcl itself.                  */
/* See also licence.terms                                         */
/* -------------------------------------------------------------- */

/* ----------------- Import interface --------------------------- */
#include <ctype.h>
#include <tcl.h>

#include "adabas.h"
#include "sqlrte.h"

/* ----------------- Export interface --------------------------- */
#include "AdabasUtil.h"

/* ----------------- Exported procedures ------------------------- */

void p2c (c_out, p_in, len)
     char *c_out;			/* Resulting null terminated string. */
     char *p_in;			/* Incoming fixed width string. */
     int   len;				/* Length of p_in. */
{
  char *pptr = p_in+len-1; /* pascal string ptr */
  char *cptr = c_out+len;  /* C      string ptr */
  
  *cptr -- = '\0';

  if (!len) return;
  
  while (pptr != p_in && * pptr == ' ') {
    *cptr-- = '\0';
    pptr--;
  }
  
  if (pptr == p_in && *pptr == ' ')
    *cptr-- = '\0';
  else {
    while (pptr != p_in)
      *cptr-- = *pptr--;
    *cptr = *pptr; /* don't forget last char */
  }
}

/* -------------------------------------------------------------- */

void c2p (p_out, c_in, len)
     char *p_out;			/* Resulting fixed width string. */
     char *c_in;			/* Incoming null terminated string. */
     int   len;				/* Length of p_out. */
{
  while (len-- > 0) {
    if (*c_in == '\0')
      *p_out++ = ' ';
    else
      *p_out++ = *c_in++;
  }
}

/* -------------------------------------------------------------- */

void escapeSequence (out, in, length)
     Tcl_DString *out;			/* Resulting escaped string. */
     char        *in;			/* Incoming data. */
     int          length;		/* Length of in. */
{
  unsigned char *x = (unsigned char *) in;
  unsigned char *p = (unsigned char *) in;
  char one_escape[5];

  while (length--) {
    if (isprint (*x) && *x != '\\')
      x++;
    else {
      if (p != x) Tcl_DStringAppend (out, (char *) p, x-p);
      sprintf (one_escape, "\\%03o", *(x++));
      Tcl_DStringAppend (out, one_escape, 4);
      p = x;
    }
  }
  if (p != x) Tcl_DStringAppend (out, (char *) p, x-p);
}

/* -------------------------------------------------------------- */

int4 substBackslash (out, in, inPos, maxLength)
     Tcl_DString *out;			/* Resulting data string. */
     char        *in;			/* Incoming escaped string. */
     int          inPos;		/* Startposition. */
     int          maxLength;		/* Length of in. */
{
  char *p   = in;
  char *old;
  int   count;
  int4  length = 0;
  char  c;

  /* First find the start position. This can't be done by pointer
   * arithmetic, since all the special symbols are escaped yet.
   */
  while (*p != 0 && inPos-- > 1) {
    if (*p == '\\') {
      c  = Tcl_Backslash (p, &count);
      p += count;
    } else
      p++;
  }

  Tcl_DStringInit (out);
  old = p;
  while (*p != 0 && (!maxLength || length < maxLength)) {
    if (*p == '\\') {
      if (p != old)
	Tcl_DStringAppend (out, old, p-old);
      c   = Tcl_Backslash (p, &count);
      Tcl_DStringAppend (out, &c, 1);
      length++;
      p  += count;
      old = p;
    } else {
      p++; length++;
    }
  }
  if (p != old)
    Tcl_DStringAppend(out, old, p-old);

  return length;
}

/* -------------------------------------------------------------- */

void hex2char (out, in, length)
     char *out;				/* returning trace like string. */
     char *in;				/* incoming data. */
     int   length;			/* length of in. */
{
  unsigned char *x = (unsigned char *) in;

  while (length-- > 0) {
    sprintf (out, "%02x", *x);
    out += 2; x++;
  }
}

/* -------------------------------------------------------------- */

int char2hex (interp, what, out, in, length)
     Tcl_Interp *interp;		/* Current interpreter. */
     char       *what;			/* specification (for errormsg). */
     char *out;				/* returning data. */
     char *in;				/* incoming trace like string. */
     int   length;			/* length of out. */
{
  char *curr = in;
  int   numb;
  unsigned char *x = (unsigned char *) out;

  while (length-- > 0) {
    if (sscanf (curr, "%2x", &numb) != 1) {
      Tcl_AppendResult (interp, "Expected ", what, ", but got \"",
			in, "\"", (char *) NULL);
      return TCL_ERROR;
    }
    *x = (unsigned char) numb;
    x++; curr += 2;
  }
  return TCL_OK;
}
