# ---------------------------------------------------------------------
#  LAST EDIT: Sun Aug 18 13:49:24 1996 by Adabas D Database (gutamil.v12-berlin.de!adabas) 
#  $Id: adabastcl.tcl,v 1.9 1996/08/18 11:52:55 adabas Exp $
# ---------------------------------------------------------------------
# Copyright (c) 1996 Christian Krone. All rights reserved.
# This program is free software; you can redistribute it and/or
# modify it under the same terms as Tcl itself.
# See also licence.terms
# ---------------------------------------------------------------------

html_start
title "Adabastcl - Adabas D Database Server access commands for Tcl"

#    ==============
ssec "Introduction"
#    ==============

p "Adabastcl is a collection of Tcl commands and a Tcl global array,
   that provides access to an Adabas D Database Server.
   Each Adabastcl command generally invokes several 
   Adabas D library functions."

p "The following section will cover all procedures of the
   [emphasis "adabastcl call interface"]. After that comes a description
   of the [func adabas] command, which provides a lower level interface
   to the database and a section about [emphasis sqlsh], a set of
   commands, that are automatically loaded by adabastclsh, if it is
   called interactive."

#    ====================
ssec "Adabastcl commands"
#    ====================

#     --------
sssec adalogon
#     --------

p [keyboard "adalogon [variable connect-str] ?[variable option] ...?"]

p "Connect to an Adabas D server using [variable connect-str].
   The connect string should be a string in one of the following forms:"

bullet_list {
  li [variable name][keyboard ,][variable password]
  li [keyboard ,][variable userkey]
  li [keyboard ,]
}

p "If a [variable userkey] is given (with a leading comma), the data for
   name and password are extracted from the xuser record.
   If only a comma is given, it stands for the DEFAULT xuser entry."

p "A logon handle is returned and should be used for all other Adabastcl
   commands using this connection that require a logon handle.  
   Multiple connections to the same or different servers are allowed.
   Adalogon raises a Tcl error if the connection 
   is not made for any reason (login or password incorrect, 
   network unavailable, etc.).  The value of the environment variable
   [path DBNAME] (which resides in the global Tcl variable
   [path env(DBNAME)] is used as the server."

p "[keyboard option] may be any of the following."

bullet_list {
  li "[keyboard -service] [variable service]"
  li "[keyboard -sqlmode] [variable sqlmode]"
}

p "If a [variable service] is given, the connection is made
   for the given service. [variable Service] may be one of the following:"

bullet_list {
  li "[keyboard user]: The connection is made for a normal user session;
      this is the default."
  li "[keyboard control]: The connection is made as user session, but with
      the privilege of the control user. Therefore the [variable connect-str]
      has to describe the control user."
  li "[keyboard utility]: The connection is made as utility session;
      Therefore the [variable connect-str] has to describe the control user.
      The created logon-handle can only be given as parameter of
      [keyboard adautil], the other commands with a [variable logon-handle]
      parameter expect a handle for a user session."
}

p "If a [variable sqlmode] is given, the connection is made in this sqlmode.
   [variable Sqlmode] can be any of [keyboard adabas], [keyboard ansi],
   [keyboard db2] or [keyboard oracle]. Default is sqlmode [keyboard adabas].
   For a utility service it is not possible to specify a sqlmode."

#     ---------
sssec adalogoff
#     ---------

p [keyboard "adalogoff [variable logon-handle]"]

p "Logoff from the Adabas D server connection associated with
   [variable logon-handle]. [variable Logon-handle] must be a valid
   handle previously opened with adalogon.
   Adalogoff returns a null string.
   Adalogoff raises a Tcl error if the logon handle specified is not open."

#     -------
sssec adaopen
#     -------

p [keyboard "adaopen [variable logon-handle]"]

p "Open an SQL cursor to the server. Adaopen returns a cursor to be used
   on subsequent Adabastcl commands that require a cursor handle.
   [variable Logon-handle] must be a valid handle previously opened with
   adalogon.
   Multiple cursors can be opened through the same or different logon
   handles, up to a maximum of 25 total cursors.
   Adaopen raises a Tcl error if the logon handle specified is not open."

#     --------
sssec adaclose
#     --------

p [keyboard "adaclose [variable cursor-handle]"]

p "Closes the cursor associated with [variable cursor-handle].
   Adaclose raises a Tcl error if the cursor handle specified is not open."

#     ------
sssec adasql
#     ------

p [keyboard "adasql [variable cursor-handle] [variable sql-statement]"]

p "Send the Adabas D SQL statement [variable sql-statement] to the server.
   [variable Cursor-handle] must be a valid handle previously opened with
   adaopen. Adasql will return the numeric return code 0 on successful
   execution of the  sql statement. The [path adamsg] array index [path rc]
   is set with the return code; the [path rows] index is set to the number
   of rows affected by the SQL statement in the case of [func insert],
   [func update], or [func delete]."

p "Only a single SQL statement may be specified in [variable sql-statement].  
   Adafetch allows retrieval of return rows generated."

p "Adasql performs an implicit [func adacancel], if any results are still
   pending from the last execution of adasql.
   Adasql raises a Tcl error if the cursor handle specified is not open, if the
   SQL statement is syntactically incorrect or if no data was found
   for a select or no row was affected by an update or delete command."

p "If the given [variable sql-statement] denotes a [emphasis "select into"],
   the selected values are stored directly into the mentioned parameter.
   The parameter names may denote ordinary variables or arrays. E.g."
preformatted {
   adasql cursor1 {SELECT DATE, TIME INTO :res(date), :res(time) FROM dual}   
}
p  "After the above call of adasql the current date and time is stored
    into the array variable [keyboard res] with the indexes [keyboard date]
    and [keyboard time]."

p "Table inserts made with adasql should follow conversion rules in
   the Adabas D SQL Reference manual."

#     --------
sssec adafetch
#     --------

p [keyboard "adafetch [variable cursor-handle] ?option...?"]

p "[keyboard option] may be any of the following."
bullet_list {
  li "[keyboard -position]   [keyboard first]"
  li "[keyboard -position]   [keyboard last]"
  li "[keyboard -position]   [keyboard next]"
  li "[keyboard -position]   [keyboard prev]"
  li "[keyboard -position]   [variable number]"
  li "[keyboard -count]      [variable mass-count]"
  li "[keyboard -command]    [variable commands]"
}

p "Return the next row from the last SQL statement executed with adasql as
   a Tcl list. [variable Cursor-handle] must be a valid handle previously
   opened with adaopen.
   Adafetch raises a Tcl error if the cursor handle specified is not open.
   All returned columns are converted to character strings.  A null string is
   returned if there are no more rows in the current set of results.
   The Tcl list that is returned by adafetch contains the values of the
   selected columns in the order specified by [func select]."

p "The cursor can be moved in any direction by means of the [func -position]
   option. Its associated value can have any of the forms [keyboard first],
   [keyboard last], [keyboard next], [keyboard prev] or it can be a
   [variable number]. The textual variants specify the direction, in which
   the cursor should be moved. A direction of e.g. [path first] rewinds
   the cursor can to the start of the result set. A [variable number]
   denotes the rownum of the result row to fetch; the first row has the
   rownum [keyboard 1]. Default position is [keyboard next]"

p "By means of the optional [variable -command] argument adafetch can
   repeatedly fetch rows and execute [variable commands] as a tcl script
   for each row.
   Substitutions are made on [variable commands] before passing it to
   [func Tcl_Eval()] for each row.  Adafetch interprets [keyboard @n]
   in commands as a result column specification.  For example,
   [keyboard "@1, @2, @3"] refer to the first, second, and third columns
   in the result. [keyboard @0] refers to the entire result row, as a Tcl
   list. Substitution columns may appear in any order, or more than once
   in the same command.
   Substituted columns are inserted into the commands string as
   proper list elements, i.e., one space will be added before and after the
   substitution and column values with embedded spaces are enclosed by
   braces if needed."

p "A Tcl error is raised if a column substitution number is greater than the
   number of columns in the results.  If the commands execute [path break],
   adafetch execution is interrupted and returns with [path TCL_OK].  Remaining
   rows may be fetched with a subsequent adafetch command.  
   If the commands execute [path return] or [path continue], the remaining
   commands are skipped and adafetch execution continues with the next row.
   Adafetch will raise a Tcl error if the \emph{commands} return an error.
   Commands should be enclosed in double quotes or braces."

p "Adatcl performs conversions for all data types.
   For raw data only a long descriptor is returned, which can be used as
   argument for a subsequent call of [func adareadlong] with its
   [keyboard -descriptor] option."

p "The [path adamsg] array index [path rc] is set with the return code
   of the fetch.  0 indicates the row was fetched successfully;
   100 indicates the end of data was reached."

p "The [path adamsg] array index [path nullvalue] can be set to specify
   the value returned when a column is null. The default is an empty
   string for values of all datatypes."

#     -------
sssec adacols
#     -------

p [keyboard "adacols [variable cursor-handle]"]

p "Return the names of the columns from the last adasql or adafetch
   command  as a Tcl list."

p "The [path adamsg] array index [path collengths] is set to a Tcl list
   corresponding to the lengths of the columns; index [path coltypes] is
   set to a Tcl list corresponding to the types of the columns;
   index [path colprecs] is set to a Tcl list corresponding to the precision
   of the numeric columns, other corresponding non-numeric columns got its
   length as precision; index [path colscales] is set to a Tcl list
   corresponding to the scale of the numeric columns,
   other corresponding non-numeric columns are 0;
   Adacols raises a Tcl error if the cursor handle specified is not open."

#     ---------
sssec adacancel
#     ---------

p [keyboard "adacancel [variable cursor-handle]"]

p "Cancels any pending results from a prior adasql command that use a
   cursor opened through the connection specified by
   [variable cursor-handle]. [variable Cursor-handle] must be a valid
   handle previously opened with [func adaopen].
   Adacancel raises a Tcl error if the cursor handle specified is not open."

#     ---------
sssec adacommit
#     ---------

p [keyboard "adacommit [variable cursor-handle]"]

p "Commit any pending transactions from prior adasql commands that use a
   cursor opened through the connection specified by logon-handle.  
   [variable Logon-handle] must be a valid handle previously opened with
   [func adalogon].
   Adacommit raises a Tcl error if the logon handle specified is not open."

#     -----------
sssec adarollback
#     -----------

p [keyboard "adarollback [variable cursor-handle]"]

p "Rollback any pending transactions from prior adasql commands that use a
   cursor opened through the connection specified by [variable logon-handle].
   [variable Logon-handle] must be a valid handle previously opened with
   adalogon.
   Adarollback raises a Tcl error if the logon handle specified is not open."

#     ----------
sssec adaautocom
#     ----------

p [keyboard "adaautocom [variable cursor-handle] [variable on-off]"]

p "Enables or disables automatic commit of SQL data manipulation 
   statements using a cursor opened through the connection specified
   by logon handle. [variable Logon-handle] must be a valid handle
   previously opened with adalogon.
   [variable on-off] must be a valid Tcl boolean value ([path 0], [path 1],
   [path false], [path true], [path no], [path yes], [path on],
   [path off] or abbreviations thereof).
   Adaautocom raises a Tcl error if the logon handle specified is not open."

#     -----------
sssec adareadlong
#     -----------

p [keyboard "adareadlong [variable cursor-handle] ?option ...?"]

p "[keyboard option] may be any of the following."
bullet_list {
  li "[keyboard -table]      [variable table-name]"
  li "[keyboard -column]     [variable column-name]"
  li "[keyboard -where]      [variable where-condition]"
  li "[keyboard -descriptor] [variable long-descriptor]"
  li "[keyboard -filename]   [variable filename]"
}

p "Read the contents of a LONG column and returns the result, or writes it
   into a file, if the optional parameter [variable filename] is specified.
   [variable Cursor-handle] must be a valid handle previously opened with
   adaopen."

p "The switches can be given in any order, but there must be present either
   the [keyboard -descriptor] switch or all of [keyboard -table],
   [keyboard -column] and [keyboard -where] switches."

p "With [keyboard -table], [keyboard -column] and [keyboard -where] the
   column to be read can be specified, as if a select in the following
   form was specified:"

p "[keyboard SELECT] [variable column-name]
   [keyboard FROM]   [variable table-name]
   [keyboard WHERE]  [variable where-condition]"

p "The given column must be of data type [path LONG] and the
   [variable where-condition] must limit the resultcount of the
   above select statement to 1."

p "[variable Long-descriptor] is the Adabas long descriptor of a
   recently fetched row, as returned by adafetch."

p "[variable Filename] is the name of a file in which to write the LONG data."

p "If called with optional parameter [variable filename], adareadlong returns
   upon successful completion the number of bytes read from the LONG column."

p "Adareadlong raises a Tcl error if the cursor handle specified is not open,
   [keyboard -descriptor] is given and either the long descriptor specified
   is invalid or no call of adafetch precedes the call of adareadlong,
   or if the [variable where-condition] doesn't limit to a single result."

#     ------------
sssec adawritelong
#     ------------

p [keyboard "adawritelong [variable cursor-handle] ?option ...?"]

p "[keyboard option] may be any of the following."
bullet_list {
  li "[keyboard -table]    [variable table-name]"
  li "[keyboard -column]   [variable column-name]"
  li "[keyboard -where]    [variable where-condition]"
  li "[keyboard -value]    [variable long-value]"
  li "[keyboard -filename] [variable filename]"
}

p "Read the contents of a file or the given string into a LONG column.
   [variable Cursor-handle] must be a valid handle previously opened with
   adaopen."

p "The switches can be given in any order, but there must be present either
   the [keyboard -value] or [keyboard -filename] switch and all of
   [keyboard -table], [keyboard -column] and [keyboard -where] switches."

p "With [keyboard -table], [keyboard -column] and [keyboard -where] the
   column to be read must be specified, as if an update in the following
   form was specified:"

p "[keyboard UPDATE] [variable table-name]
   [keyboard SET]    [variable column-name] = '[variable value-to-be-inserted]'
   [keyboard WHERE]  [variable where-condition]"

p "The given column must be of data type [path LONG] and the
   [variable where-condition] must limit the resultcount of the
   above update statement to 1."

p "[variable Filename] is the name of a file out of which to read the
   LONG data. [variable Long-value] is the LONG data itself."

p "Adawritelong raises a Tcl error if the cursor handle specified is not open
   or if the [variable where-condition] doesn't limit to a single result."

#     -------
sssec adautil
#     -------

p [keyboard "adautil [variable logon-handle] [variable utility-command]"]

p "Send the Adabas D utility command [variable utility-command] to the server.
   [variable Logon-handle] must be a valid handle previously opened with
   adalogon with the session specified as [keyboard utility].
   The return value depends heavily on the specified command."

#     --------------------------------------
sssec "Server message and error information"
#     --------------------------------------

p "Adabastcl creates and maintains a Tcl global array to provide feedback of
   Adabas D server messages, named [path adamsg]. [path Adamsg] is also used
   to communicate with the Adabastcl interface routines to
   specify null return values.
   In all cases except for [path nullvalue] and [path version],
   the contents of each element may be changed
   upon invocation of any Adabastcl command, and any 
   element affected by the command is set. The [path adamsg] array is shared 
   among all open Adabastcl handles.  [path Adamsg] should be defined with the
   global statement in any Tcl procedure needing access to [path adamsg].
   The following list defines all indexes of [path adamsg]."

definition_list {
  term [path nullvalue]
  term_definition "can be set by the programmer to indicate the string value
                   returned for any null result. Setting
                   [path adamsg(nullvalue)] to an empty string or unsetting
                   it will return an empty string for all null values.
                   [path Nullvalue] is initially set to an empty string."
  term [path handle]
  term_definition "indicates the handle of the last Adabastcl command.
                   [path Handle] is set on every Adabastcl command (execpt
                   where an invalid handle is used.)"
  term [path rc]
  term_definition "indicates the results of the last SQL command and
                   subsequent adafetch processing. [path Rc] is set by
                   [func adasql], [func adafetch], and is the numeric
                   return code from the last library function called by
                   an Adabastcl command.  Refer to Adabas D Error Messages
                   and Codes manual for detailed information."

                p "Typical values are:"
                   bullet_list {
                     li "0:   Function completed normally, without error."
                     li "100: End of data was reached on an adafetch command,
                              no data was found for a select on a adasql
                              command or no row was affected by an update
                              or delete command on a adasql command."
                     li "all other returncodes unequal to 0 or 100:
                              invalid SQL statement, invalid sql statements,
                              missing keywords, invalid column names, no sql
                              statement, logon denied, insufficient
                              privileges, etc. The returncode corresponds
                              to the number in the messages manual of
                              Adabas D."
                   }
  term [path errortxt]
  term_definition "the message text associated with [path rc]."
  term [path collengths]
  term_definition "is a Tcl list of the lengths of the columns returned by
                   adacols. Collengths is only set by adacols."
  term [path coltypes]
  term_definition "is a Tcl list of the types of the columns returned by
                   [func adacols]. [path Coltypes] is only set by adacols.
                   Possible types returned are:
                   [path fixed], [path float], [path char_ascii],
                   [path char_ebcdic], [path char_byte], [path rowid],
                   [path long_ascii], [path long_ebcdic], [path long_byte],
                   [path long_dbyte], [path long_unicode], [path date],
                   [path time], [path vfloat], [path timestamp],
                   [path duration], [path dbyte_ebcdic], [path boolean],
                   [path unicode], [path smallint], [path integer],
                   [path varchar_ascii], [path varchar_ebcdic],
                   [path varchar_byte] or [path unknown]"
  term [path colprecs]
  term_definition "is a Tcl list of the precision of the numeric columns
                   returned by adacols. Colprecs is only set by adacols.
                   For non-numeric columns, the list entry is a zero."
  term [path colscales]
  term_definition "is a Tcl list of the scale of the numeric columns returned
                   by adacols. Colprecs is only set by adacols.
                   For non-numeric columns, the list entry is a zero."
  term [path rows]
  term_definition "the number of rows affected by an [func insert],
                   [func update], or [func delete] in an adasql command."
  term [path version]
  term_definition "a string containing the version of Adabastcl and the
                   version of the Adabas D server, it is compiled for, in
                   the form of e.g.
                   [screen "AdabasTcl 1.2 (for Adabas D 6.2)"]."
}

#    ====================
ssec "The adabas command"
#    ====================

p "The [func adabas] command provides a low level interface to the
   Adabas D database server. In deed, it is so low level, that you must
   know some internals about the order interface between applications and
   kernel, to get the most of the command variants functional."

p "In theory, it is possible to rebuild all the above given commands of
   the adabastcl call interface as purely [prog Tcl] commands by means of
   this command. But its performance is very bad, since all data at this
   level is binary, and [prog Tcl] isn't very good in handling strings
   with null characters in it."

p "But some options of this command are nevertheless convenient, e.g.
   [keyboard "adabas xuser"] or [keyboard "adabas crypt"]."

sssec "adabas version"

sssec "adabas crypt"
sssec "adabas termId"

sssec "adabas get"
sssec "adabas put"

sssec "adabas fget"
sssec "adabas fapp"
sssec "adabas sget"

sssec "adabas xuser"
bullet_list {
  li "[keyboard args]"
  li "[keyboard clear]"
  li "[keyboard close]"
  li "[keyboard get]"
  li "[keyboard index]"
  li "[keyboard open]"
  li "[keyboard put]"
}

sssec "adabas open"

#    =====================================
ssec "A SQL shell for interactive queries"
#    =====================================

p "If [prog adabastclsh] was called without a filename to source, and
   therefore [keyboard tcl_interactive] is set to 1, it reads at startup
   time a file called [path ~/.adabastclshrc]. In this procedure you
   can e.g. set your prompt. It is recommended put the following three
   line into this file:"

preformatted {
  if {[catch {source $env(DBROOT)/tcl/sqlsh.tcl} msg]} {
    puts "Error while initializing adabastclsh:\n$msg"
  }
}

p "This file defines a bunch of convinience commands, that have
   the names of all the Adabas SQL commands."

bullet_list {
  li alter
  li clear
  li commit
  li comment
  li connect
  li create
  li delete
  li drop
  li exists
  li explain
  li grant
  li insert
  li load
  li monitor
  li refresh
  li revoke
  li rollback
  li select
  li show
  li unload
  li update
  li rename
}

p "The Tcl commands [func rename] and [func update] are accessible as
   [func tcl_rename] and [func tcl_update]."

p "The [func connect] opens a connection to the database, that keeps
   established, until a session timeout occured or a [func commit] or
   [func release] with the [keyboard release] option was given."

p "After a [func select] statement, that returns no error and contains no
   [keyboard into] clause, all results are fetched in portions of 25 lines.
   At the prompt the user can quit the fetch by typing [keyboard q], can
   scroll the remaining lines without further prompting by typing
   [keyboard n] or can see the next lines by typing any other key."

p "All other commands are just send to the database kernel and the
   resulting errormessage, if any, is returned as error."
 
p "So an example session could look like the following."

preformatted {
  connect krischan geheim
  select date, time into :x, :y from dual
  select * from order where order_date = '$x'
  delete from account where account_date <> '$x'
  commit work release
}

p "Beside the above given SQL commands there are two more commands defined:
   [func utility] to connect as control user on the utility service and
   [func util], to perform some commands with this connection.
   Again an example:"

preformatted {
  utility control control
  util state
  util diagnose tabid krischan.adresse
  util commit release
}
  
#    =======
ssec "Notes"
#    =======

p "Tcl errors can also be raised by any Adabastcl command if a command's
   internal calls to library routines fail."

#    =======================
ssec "Environment variables"
#    =======================

definition_list {
  term [path DBNAME]
  term_definition "The default Adabas D server name."
  term [path ADABASTCL_DIR]
  term_definition "The directory to look for the startup scripts;
                   also the init value for the global Tcl variable
                   [path adabastcl_library]."
}

#    =======
ssec "Files"
#    =======

definition_list {
  term [path {$HOME/.XUSER}]
  term_definition or
  term [path {$HOME/.XUSER.62}]
  term_definition "The xuser file to look for connect parameters."
}

#    ======
ssec "Bugs"
#    ======

bullet_list {
  li "non known so far."
}

#    =========
ssec "Credits"
#    =========

p "Adabastcl is deeply inspired by [path orasql] (SQL interface to Oracle
    server) by Tom Poindexter"


#    ========
ssec "Author"
#    ========

p "Christian Krone, Berlin. Version 0.4, August 1996."
p [path krischan@cs.tu-berlin.de]

html_end

