# -----------------------------------------------------------------------
#  LAST EDIT: Sat Aug 17 19:05:53 1996 by Adabas D Database (gutamil.v12-berlin.de!adabas) 
#  $Id: tkquery.tcl,v 1.6 1996/08/17 17:06:03 adabas Exp $
# -----------------------------------------------------------------------
# Copyright (c) 1996 Christian Krone. All rights reserved.
# This program is free software; you can redistribute it and/or
# modify it under the same terms as Tcl itself.
# See also licence.terms
# -----------------------------------------------------------------------

html_start
title "Tkquery - A tool for interactive SQL queries"

#    ----------
ssec "Synopsis"
#    ----------

p [path "tkquery \[-n [variable dbnode]\] \[-d [variable dbname]\] \
                 \[-u [variable user],[variable passwd]\] \
                 \[-U [variable xuserKey]\]"]

#    ---------
ssec "Options"
#    ---------

definition_list {
  term [path -n]
  term_definition "database node (default=local);"
  term [path  -d]
  term_definition "database name (default=DBNAME);"
  term [path  -u]
  term_definition "name and password of user;"
  term [path  -U]
  term_definition "name of xuser key."
}

#    --------------
ssec "Introduction"
#    --------------

p "When starting up the application, the user has to give the connect data
   by filling in the forms of the connect window or using the command line
   options. If every thing is ok, a window called TkQuery is displayed."

p "The window consists mainly of four parts, which are described here from
   top to bottom."

#    -----------
sssec "Menu bar"
#    -----------

p "By means of the menu bar the user can start all the functionality,
   that tkquery offers. The most important can be called also by the
   press of a function key; for the list of possible keys see
   [ref ssec "Keyboard shortcuts"] below."

p "The top most entries of the [func SQL] menu are for navigate through
   the history of your sql queries. [func New] clears the SQL window,
   while [func Prev] and [func Next] cycles through the history. How
   many queries are stored, can be customized with the resource
   [path history] (see [ref ssec X-Resources])."

p "In the [func SQL] menu there are four entries to store and retrieve
   the queries in the database (beneath the title
   [screen "Stored commands"]). These functions store the queries in the
   same system tables and in the same format as [prog query] (the classic
   query tool of Adabas D) does; so you can exchange queries between
   these both tools via the database."

bullet_list {
  li "[func Get...] pops up a scollable list containing the names of all
      stored commands. You can select a query by moving the active element
      with your up and down arrow keys or by clicking on the name of a
      query with your mouse pointer. A click on [func Okay] or a double
      click of the left mouse button starts the reading of the selected
      query in the database, which would then be displayed in the SQL window."

  li "[func Replace] only works, if you already connected your query with
      a stored command by [func Get...] or [func Insert...]. Then it
      updates the query in the system tables with the query, that is
      displayed in the SQL window."

  li "[func Insert...] pops up a dialog with an text entry, where you can
      determine the name of the query to store."

  li "[func Delete...] pops up a list similar to [func Get...], and after
      selecting a stored command it is deleted in the system tables."
}

p "The three entries [func Open...], [func Save] and [func "Save as..."]
   are to store and retrieve the queries from the local file system.
   They pop up a file selection box, where the user can give a filename."

p "The bottom most entry in [func SQL] menu is for quitting the program."

p "The menu [func Results] is for handling of the result window
   (see [ref sssec "Results window"]).
   Here you can clear the result window ([func Clear])
   or save its content into a file ([func Save...]).
   You can activate the protocol, then every query together with its
   results are appended to the file [path tkquery.prot]."

p "The output depends on the [func Style] cascade menu of the
   [func Options] menu. It can be a table in plain text or in HTML
   or LaTeX format. If the output style is plain, you can call
   [func Preview...] and a window with the results formatted in
   plain style will pop up. If the output style is HTML and
   [prog netscape] is running on the same machine and display as
   [prog tkquery], [prog netscape] will display the results."

p "The [func Options] menu is mostly for customize personal preferences.
   All options can be configured via resources at startup time of
   [prog tkquery] (see [ref ssec X-Resources])."

p "The two entries [func "Commit now"] and [func "Rollback now"] are
   not really options; they stay here just because its relations to
   the [func Autocommit] options. If autocommit is disabled, these
   both commands commit (or rollback) all the sql commands entered since
   the last call of these commands or since the disabling of autocommit."

p "In the menu [func "DB objects"] you can look at your own or on all
   accessible data base objects (tables, columns, views and indexes).
   After clicking on a menu entry a window pops up, that displays the
   wanted objects.
   If you click on a column in a window showing tables, another window with
   information about the columns of this table will pop up.
   If you click on an index in a window showing indexes, another window with
   information about the columns of this index will pop up."

#    -------------
sssec "Sql window"
#    -------------

p "In this window the user can enter a SQL statement.
   No special formatting is neccessary (Layout will be ignored).
   The motif or emacs like control sequences are activ (e.g. [keyboard C-a]
   for move to beginning of line). The user can start the entered query by
   a click on the [screen Execute] button or by a press of the
   [keyboard F5] key."

p "If the SQL statement is syntactically incorrect, the insert cursor is
   placed at the point, where the database server detected the error."

p "The user can scroll the SQL window with the attached scrollbar, can clear
   it and even switch to the previos or next entered SQL statement."

#    -----------------
sssec "Results window"
#    -----------------

p "The results of an executed select statement are displayed in this window.
   If there are more results than fit into the window, you can scroll
   in every direction with the scrollbars, that are attached to the right
   and bottom of the window (or with [keyboard "Page Up"],
   [keyboard "Page Down"], [keyboard F10] and [keyboard F11])."

p "You can exclude a column from displaying by clicking its name with the
   right mouse button. This is a shortcut for the [func Configure...]
   entrie in the [func Results] menu (see [ref sssec "Menu bar"]).
   There you can reinclude a column, too."

p "The results can be printed together with the query into a file by
   means of the function [func Print...] in the menu [func Results].
   All queries together with their results are printed in a file called
   [path tkquery.prot], if the checkbutton [func Protcol] is activated
   (the square to its left is red).
   The format of the printing, which can be either plain text, latex or html,
   can be selected in the [func Options] menu."

p "Also in the [func Options] menu there is the possibility to select a font
   for the results window."

#    --------------
sssec "Status line"
#    --------------

p "In this line there is the friendly message [screen "Enter command"] or
   the message returned from the last executed command."

#    --------------------
ssec "Keyboard shortcuts"
#    --------------------

definition_list {
  term [keyboard F1]
  term_definition "A window with this help text pops up."
  term [keyboard F2]
  term_definition "The SQL window and the result window will be cleared."
  term [keyboard F3]
  term_definition "If currently a query is in work, it will be cancelled and
                   tkquery will fetch no more additional rows. If no query
                   is at work, the program will be terminated."
  term [keyboard F5]
  term_definition "The query in the SQL window will be executed, and in the
                   case of a SELECT statement, the resulting rows are
                   displayed in the result window."
  term [keyboard F6]
  term_definition "The next SQL statement will be shown in the SQL window."
  term [keyboard F9]
  term_definition "The previous SQL statement will be shown in the SQL window."
  term [keyboard F10]
  term_definition "The result window will scroll a page to the left,
                   if possible."
  term [keyboard F11]
  term_definition "The result window will scroll a page to the right,
                   if possible."
  term "Movement keys"
  term_definition "The input cursor of the SQL window will move in the
                   corresponding direction."
  term "[keyboard "Page Up"], [keyboard "Page down"]"
  term_definition "The result window will scroll up (or down) a page,
                   if possible."
}

#    -------
ssec "Notes"
#    -------

p "In stored commands created by [prog query] there can be multiple
   sql statements, each followed by a list of report commands.
   [prog TkQuery] executes only the first query of a stored command
   and ignores all the report statement at its end. Also the parameter
   mechanism (with % or &) is ignored by [prog TkQuery]."

#    -----------------------
ssec "Environment variables"
#    -----------------------

definition_list {
  term [path DBNAME]
  term_definition "The default Adabas D server name."
  term [path ADABASTCL_DIR]
  term_definition "The directory to look for the startup scripts;
                   also the init value for the global Tcl variable
                   [path adabastcl_library]."
}

#    -------------
ssec "X-Resources"
#    -------------

p "While starting, the X resource database is read, so that the startup
   state of some check or radiobuttons can be custumized by the user."

p "The resources can be set via the [prog xrdb] command (highest priority)
   or mentioned in the file [path \$HOME/.Xdefaults].
   The default values, if no resource assignment can be found,
   can bee seen in the following excerpt from a resource file."

preformatted {
  *tkquery.fontName:     helvetica
  *tkquery.fontSize:     12
  *tkquery.fontStyle:    bold
  *tkquery.autoCommit:   False
  *tkquery.dateFormat:   internal
  *tkquery.outputStyle:  plain
  *tkquery.withProt:     False
  *tkquery.history:      10
  *tkquery.nullValue:
  *tkquery.netscapeHelp: False
}

#    -------
ssec "Files"
#    -------

definition_list {
  term [path {$HOME/.XUSER}]
  term_definition or
  term [path {$HOME/.XUSER.62}]
  term_definition "The xuser file to look for connect parameters."
}

#    ------
ssec "Bugs"
#    ------

bullet_list {
  li "There should be an option for the sqlmode."
}

#    ---------
ssec "Credits"
#    ---------

p "Tkquery is deeply inspired by and build upon [path wosql] 
   (Windowing Oracle SQL) by Tom Poindexter"

#    --------
ssec "Author"
#    --------

p "Christian Krone, Berlin. Version 0.4, August 1996."
p [path krischan@cs.tu-berlin.de]

html_end
